@@ -2,7 +2,7 @@
 // Syd: rock-solid application kernel
 // src/compat.rs: Compatibility code for different libcs
 //
-// Copyright (c) 2023, 2024, 2025 Ali Polatel <alip@chesswob.org>
+// Copyright (c) 2023, 2024, 2025, 2026 Ali Polatel <alip@chesswob.org>
 // waitid and WaitStatus are based in part upon nix which is:
 //     Copyright (c) nix authors
 //     SPDX-License-Identifier: MIT
@@ -10,25 +10,32 @@
 // SPDX-License-Identifier: GPL-3.0
 
 #![allow(non_camel_case_types)]
+#![allow(non_upper_case_globals)]
 
 use std::{
     alloc::{alloc, dealloc, Layout},
     cell::RefCell,
     cmp::Ordering,
-    os::fd::{AsFd, AsRawFd, RawFd},
+    ffi::CStr,
+    fmt,
+    num::NonZeroUsize,
+    os::fd::{AsFd, AsRawFd, BorrowedFd, FromRawFd, OwnedFd, RawFd},
     ptr::NonNull,
     sync::LazyLock,
 };
 
 use bitflags::bitflags;
 use libseccomp::ScmpSyscall;
+use memchr::arch::all::is_equal;
 use nix::{
     errno::Errno,
     fcntl::{AtFlags, OFlag},
     sys::{
         epoll::EpollOp,
+        signal::Signal,
         socket::SockaddrLike,
-        stat::makedev,
+        stat::Mode,
+        time::TimeSpec,
         wait::{Id, WaitPidFlag, WaitStatus as NixWaitStatus},
     },
     unistd::Pid,
@@ -36,10 +43,79 @@ use nix::{
 };
 use serde::{ser::SerializeMap, Serialize, Serializer};
 
-use crate::{config::*, lookup::FileType, XPath};
+use crate::{
+    config::*, ioctl::Ioctl, lookup::FileType, parsers::sandbox::str2u64, retry::retry_on_eintr,
+    XPath,
+};
+
+#[repr(C)]
+#[derive(Copy, Clone)]
+#[allow(dead_code)]
+struct syscall_ll_pair(libc::c_ulong, libc::c_ulong);
+
+// Split a 64-bit value into two 32-bit values ordered by endianness.
+//
+// This macro mimics musl's __SYSCALL_LL_E macro.
+#[allow(unused_macros)]
+macro_rules! syscall_ll_e {
+    ($val:expr) => {{
+        // All 32-bit except x32 and n32.
+        #[cfg(all(
+            target_pointer_width = "32",
+            not(target_arch = "x86_64"),
+            not(target_arch = "mips64")
+        ))]
+        {
+            let val = $val as libc::c_ulonglong;
+            let lo = (val & 0xffff_ffff) as libc::c_ulong;
+            let hi = (val >> 32) as libc::c_ulong;
+            if cfg!(target_endian = "little") {
+                syscall_ll_pair(lo, hi)
+            } else {
+                syscall_ll_pair(hi, lo)
+            }
+        }
+        #[cfg(not(all(
+            target_pointer_width = "32",
+            not(target_arch = "x86_64"),
+            not(target_arch = "mips64")
+        )))]
+        {
+            $val
+        }
+    }};
+}
+
+// AT_EACCESS is not defined in nix for Android, but it is 0x200 on Linux.
+pub(crate) const AT_EACCESS: nix::fcntl::AtFlags = nix::fcntl::AtFlags::from_bits_retain(0x200);
 
 pub(crate) const UIO_MAXIOV: usize = libc::UIO_MAXIOV as usize;
-pub(crate) const UIO_MAXLEN: usize = 256_000;
+
+// IPC_SET is 1 on Linux.
+pub(crate) const IPC_SET: i32 = 1;
+
+/// On Linux, we use the libc definitions for ptrace(2).
+/// On Android, these are missing from libc, so we define them.
+#[cfg(not(target_os = "android"))]
+pub use libc::{NT_PRSTATUS, PTRACE_LISTEN, PTRACE_SEIZE};
+
+#[cfg(target_os = "android")]
+pub const NT_PRSTATUS: libc::c_int = 1;
+#[cfg(target_os = "android")]
+pub const PTRACE_SEIZE: libc::c_int = 0x4206;
+#[cfg(target_os = "android")]
+pub const PTRACE_LISTEN: libc::c_int = 0x4208;
+
+// SYS_fstatfs is 44 on aarch64, 100 on arm/x86, and 138 on x86_64.
+#[cfg(all(target_os = "android", target_arch = "aarch64"))]
+pub(crate) const SYS_fstatfs: libc::c_long = 44;
+#[cfg(all(target_os = "android", any(target_arch = "arm", target_arch = "x86")))]
+pub(crate) const SYS_fstatfs: libc::c_long = 100;
+#[cfg(all(target_os = "android", target_arch = "x86_64"))]
+pub(crate) const SYS_fstatfs: libc::c_long = 138;
+// On Linux, use libc definitions.
+#[cfg(not(target_os = "android"))]
+pub(crate) use libc::SYS_fstatfs;
 
 // x32 compatibility
 // See https://sourceware.org/bugzilla/show_bug.cgi?id=16437
@@ -186,6 +262,9 @@ impl From<libc::stat64> for stat32 {
             __pad0: 0,
             st_rdev: stat.st_rdev as u64,
             st_size: stat.st_size as i32,
+            #[cfg(target_os = "android")]
+            st_blksize: stat.st_blksize as u64,
+            #[cfg(not(target_os = "android"))]
             st_blksize: stat.st_blksize,
             st_blocks: stat.st_blocks as i32,
             st_atime: stat.st_atime as i32,
@@ -859,8 +938,8 @@ impl Drop for DirIter {
     }
 }
 
-impl std::fmt::Debug for DirIter {
-    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
+impl fmt::Debug for DirIter {
+    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
         f.debug_struct("DirIter")
             .field("bufsiz", &self.bufsiz)
             .field("memsiz", &self.memsiz)
@@ -879,8 +958,8 @@ pub struct DirEntry<'a> {
     namelen: usize,
 }
 
-impl std::fmt::Debug for DirEntry<'_> {
-    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
+impl fmt::Debug for DirEntry<'_> {
+    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
         f.debug_tuple("DirEntry")
             .field(&self.as_xpath())
             .field(&self.file_type())
@@ -1082,17 +1161,17 @@ pub enum WaitStatus {
 impl From<NixWaitStatus> for WaitStatus {
     fn from(status: NixWaitStatus) -> Self {
         match status {
-            NixWaitStatus::Exited(pid, code) => WaitStatus::Exited(pid, code),
+            NixWaitStatus::Exited(pid, code) => Self::Exited(pid, code),
             NixWaitStatus::Signaled(pid, signal, core_dump) => {
-                WaitStatus::Signaled(pid, signal as i32, core_dump)
+                Self::Signaled(pid, signal as i32, core_dump)
             }
-            NixWaitStatus::Stopped(pid, signal) => WaitStatus::Stopped(pid, signal as i32),
+            NixWaitStatus::Stopped(pid, signal) => Self::Stopped(pid, signal as i32),
             NixWaitStatus::PtraceEvent(pid, signal, event) => {
                 WaitStatus::PtraceEvent(pid, signal as i32, event)
             }
-            NixWaitStatus::PtraceSyscall(pid) => WaitStatus::PtraceSyscall(pid),
-            NixWaitStatus::Continued(pid) => WaitStatus::Continued(pid),
-            NixWaitStatus::StillAlive => WaitStatus::StillAlive,
+            NixWaitStatus::PtraceSyscall(pid) => Self::PtraceSyscall(pid),
+            NixWaitStatus::Continued(pid) => Self::Continued(pid),
+            NixWaitStatus::StillAlive => Self::StillAlive,
         }
     }
 }
@@ -1231,6 +1310,54 @@ pub(crate) fn pipe2_raw(flags: OFlag) -> Result<(RawFd, RawFd), Errno> {
     Ok((read, write))
 }
 
+// sigwaitinfo(2) is not exported by bionic on Android.
+// Implement it using sigtimedwait(2) with a NULL timeout.
+pub(crate) unsafe fn sigwaitinfo(
+    set: *const libc::sigset_t,
+    info: *mut libc::siginfo_t,
+) -> libc::c_int {
+    libc::sigtimedwait(set, info, std::ptr::null())
+}
+
+// timer_create(2) is not exported by bionic on Android.
+#[expect(clippy::cast_possible_truncation)]
+#[expect(clippy::cast_sign_loss)]
+pub(crate) unsafe fn timer_create(
+    clockid: libc::clockid_t,
+    sevp: *mut libc::sigevent,
+    timerid: *mut libc::timer_t,
+) -> libc::c_int {
+    let mut kernel_timer_id: libc::c_int = 0;
+    let res = libc::syscall(libc::SYS_timer_create, clockid, sevp, &mut kernel_timer_id);
+    if res == 0 {
+        *timerid = kernel_timer_id as usize as libc::timer_t;
+    }
+    res as libc::c_int
+}
+
+// timer_settime(2) is not exported by bionic on Android.
+#[expect(clippy::cast_possible_truncation)]
+pub(crate) unsafe fn timer_settime(
+    timerid: libc::timer_t,
+    flags: libc::c_int,
+    new_value: *const libc::itimerspec,
+    old_value: *mut libc::itimerspec,
+) -> libc::c_int {
+    libc::syscall(
+        libc::SYS_timer_settime,
+        timerid,
+        flags,
+        new_value,
+        old_value,
+    ) as libc::c_int
+}
+
+// timer_delete(2) is not exported by bionic on Android.
+#[expect(clippy::cast_possible_truncation)]
+pub(crate) unsafe fn timer_delete(timerid: libc::timer_t) -> libc::c_int {
+    libc::syscall(libc::SYS_timer_delete, timerid) as libc::c_int
+}
+
 #[expect(clippy::cast_possible_truncation)]
 pub(crate) const PF_UNSPEC: libc::sa_family_t = libc::AF_UNSPEC as libc::sa_family_t;
 #[expect(clippy::cast_possible_truncation)]
@@ -1311,7 +1438,7 @@ pub fn statx<Fd: AsFd, P: ?Sized + NixPath>(
     mask: libc::c_uint,
 ) -> Result<FileStatx, Errno> {
     let dirfd = dirfd.as_fd().as_raw_fd();
-    let mut dst = std::mem::MaybeUninit::uninit();
+    let mut dst = std::mem::MaybeUninit::zeroed();
 
     // SAFETY: Force DONT_SYNC if FORCE_SYNC is not set.
     if flags & AT_STATX_FORCE_SYNC == 0 {
@@ -1339,7 +1466,7 @@ pub fn statx<Fd: AsFd, P: ?Sized + NixPath>(
 /// This function always sets the flag AT_STATX_DONT_SYNC.
 pub fn fstatx<Fd: AsFd>(fd: Fd, mask: libc::c_uint) -> Result<FileStatx, Errno> {
     let fd = fd.as_fd().as_raw_fd();
-    let mut dst = std::mem::MaybeUninit::uninit();
+    let mut dst = std::mem::MaybeUninit::zeroed();
 
     // SAFETY: Neither nix nor libc has a wrapper for statx.
     Errno::result(unsafe {
@@ -1410,17 +1537,17 @@ pub(crate) fn fstatat64<Fd: AsFd, P: ?Sized + NixPath>(
 }
 
 #[cfg(target_os = "freebsd")]
-type fs_type_t = u32;
+pub(crate) type fs_type_t = u32;
 #[cfg(target_os = "android")]
-type fs_type_t = libc::c_ulong;
+pub(crate) type fs_type_t = libc::c_ulong;
 #[cfg(all(target_os = "linux", target_arch = "s390x", not(target_env = "musl")))]
-type fs_type_t = libc::c_uint;
+pub(crate) type fs_type_t = libc::c_uint;
 #[cfg(all(target_os = "linux", target_env = "musl"))]
-type fs_type_t = libc::c_ulong;
+pub(crate) type fs_type_t = libc::c_ulong;
 #[cfg(all(target_os = "linux", target_env = "ohos"))]
-type fs_type_t = libc::c_ulong;
+pub(crate) type fs_type_t = libc::c_ulong;
 #[cfg(all(target_os = "linux", target_env = "uclibc"))]
-type fs_type_t = libc::c_int;
+pub(crate) type fs_type_t = libc::c_int;
 #[cfg(all(
     target_os = "linux",
     not(any(
@@ -1430,17 +1557,40 @@ type fs_type_t = libc::c_int;
         target_env = "uclibc"
     ))
 ))]
-type fs_type_t = libc::__fsword_t;
+pub(crate) type fs_type_t = libc::__fsword_t;
 
-const BTRFS_SUPER_MAGIC: fs_type_t = libc::BTRFS_SUPER_MAGIC as fs_type_t;
-const HUGETLBFS_MAGIC: fs_type_t = libc::HUGETLBFS_MAGIC as fs_type_t;
-const OVERLAYFS_SUPER_MAGIC: fs_type_t = libc::OVERLAYFS_SUPER_MAGIC as fs_type_t;
-const PROC_SUPER_MAGIC: fs_type_t = libc::PROC_SUPER_MAGIC as fs_type_t;
+/// Filesystem type
+#[derive(Clone, Copy, Debug, Eq, PartialEq, Ord, PartialOrd, Hash)]
+pub struct FsType(fs_type_t);
 
-/// Wrapper for struct statfs64
-pub(crate) struct Statfs64(libc::statfs64);
+impl FsType {
+    /// Fetches the filesystem type of the given file descriptor.
+    pub fn get<Fd: AsFd>(fd: Fd) -> Result<Self, Errno> {
+        retry_on_eintr(|| fstatfs64(&fd)).map(|stfs| Self(stfs.0.f_type))
+    }
+
+    /// Resolves a filesystem name to a list of `FsType`s.
+    pub fn from_name(name: &str) -> Result<Vec<Self>, Errno> {
+        // Parse as integer first, and as name next.
+        let mut fs_types = Vec::new();
+
+        if let Ok(fs_type) = str2u64(name.as_bytes()) {
+            #[expect(clippy::cast_possible_wrap)]
+            fs_types.push(Self(fs_type as fs_type_t));
+        } else {
+            for (fs_name, fs_type) in FS_MAGIC {
+                if is_equal(name.as_bytes(), fs_name.as_bytes()) {
+                    fs_types.push(fs_type.into());
+                }
+            }
+        }
+        if fs_types.is_empty() {
+            return Err(Errno::EINVAL);
+        }
+
+        Ok(fs_types)
+    }
 
-impl Statfs64 {
     // WORKAROUND:
     // Check if the file resides on a btrfs|overlayfs.
     // Overlayfs does not report device IDs correctly on
@@ -1449,18 +1599,101 @@ impl Statfs64 {
     // Btrfs has the same issue:
     // https://www.reddit.com/r/btrfs/comments/1clgd8u/different_dev_id_reported_by_statx_and/
     // Assume true on errors for safety.
-    pub(crate) fn has_broken_device_ids(&self) -> bool {
-        matches!(self.0.f_type, OVERLAYFS_SUPER_MAGIC | BTRFS_SUPER_MAGIC)
+    pub(crate) fn has_broken_devid(self) -> bool {
+        self.is_overlayfs() || self.is_btrfs()
     }
 
-    // Check if file resides on a hugetlbfs.
-    pub(crate) fn is_huge_file(&self) -> bool {
-        self.0.f_type == HUGETLBFS_MAGIC
+    /// Check if file resides on a hugetlbfs.
+    pub fn is_huge_file(self) -> bool {
+        self.0 == HUGETLBFS_MAGIC
     }
 
-    // Check if file resides on a procfs.
-    pub(crate) fn is_proc(&self) -> bool {
-        self.0.f_type == PROC_SUPER_MAGIC
+    /// Check if file resides on a procfs.
+    pub fn is_proc(self) -> bool {
+        self.0 == PROC_SUPER_MAGIC
+    }
+
+    /// Check if file resides on an overlayfs.
+    pub fn is_overlayfs(self) -> bool {
+        self.0 == OVERLAYFS_SUPER_MAGIC
+    }
+
+    /// Check if file resides on a btrfs.
+    pub fn is_btrfs(self) -> bool {
+        self.0 == BTRFS_SUPER_MAGIC
+    }
+
+    /// Check if file resides on a zfs.
+    pub fn is_zfs(self) -> bool {
+        self.0 == ZFS_SUPER_MAGIC
+    }
+}
+
+impl fmt::Display for FsType {
+    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
+        let magic = self.0;
+
+        if let Some((name, _)) = FS_MAGIC.iter().find(|(_, m)| *m == magic) {
+            f.write_str(name)
+        } else {
+            write!(f, "{magic:#x}")
+        }
+    }
+}
+
+impl Serialize for FsType {
+    fn serialize<S>(&self, serializer: S) -> Result<S::Ok, S::Error>
+    where
+        S: Serializer,
+    {
+        let magic = self.0;
+
+        if let Some((name, _)) = FS_MAGIC.iter().find(|(_, m)| *m == magic) {
+            serializer.serialize_str(name)
+        } else {
+            #[expect(clippy::cast_sign_loss)]
+            serializer.serialize_u64(magic as u64)
+        }
+    }
+}
+
+impl From<fs_type_t> for FsType {
+    fn from(fs_type: fs_type_t) -> Self {
+        FsType(fs_type)
+    }
+}
+
+impl From<FsType> for fs_type_t {
+    fn from(fs_type: FsType) -> Self {
+        fs_type.0
+    }
+}
+
+impl From<&fs_type_t> for FsType {
+    fn from(fs_type: &fs_type_t) -> Self {
+        FsType(*fs_type)
+    }
+}
+
+impl From<&FsType> for fs_type_t {
+    fn from(fs_type: &FsType) -> Self {
+        fs_type.0
+    }
+}
+
+const BTRFS_SUPER_MAGIC: fs_type_t = libc::BTRFS_SUPER_MAGIC as fs_type_t;
+const HUGETLBFS_MAGIC: fs_type_t = libc::HUGETLBFS_MAGIC as fs_type_t;
+const OVERLAYFS_SUPER_MAGIC: fs_type_t = libc::OVERLAYFS_SUPER_MAGIC as fs_type_t;
+const PROC_SUPER_MAGIC: fs_type_t = libc::PROC_SUPER_MAGIC as fs_type_t;
+const ZFS_SUPER_MAGIC: fs_type_t = 0x2fc12fc1i64 as fs_type_t;
+
+/// Wrapper for struct statfs64
+pub struct Statfs64(libc::statfs64);
+
+impl Statfs64 {
+    /// Returns filesystem type.
+    pub fn fs_type(&self) -> FsType {
+        FsType(self.0.f_type)
     }
 }
 
@@ -1683,8 +1916,8 @@ impl From<u64> for LsmId {
     }
 }
 
-impl std::fmt::Display for LsmId {
-    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
+impl fmt::Display for LsmId {
+    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
         match self {
             LsmId::Undef => write!(f, "undef"),
             LsmId::Capability => write!(f, "capability"),
@@ -1991,6 +2224,12 @@ bitflags! {
         /// Use hugetlbfs.
         const MFD_HUGETLB = libc::MFD_HUGETLB;
 
+        /// Huge page size: 64KB
+        const MFD_HUGE_64KB = libc::MFD_HUGE_64KB;
+
+        /// Huge page size: 512KB
+        const MFD_HUGE_512KB = libc::MFD_HUGE_512KB;
+
         /// Huge page size: 1MB.
         const MFD_HUGE_1MB = libc::MFD_HUGE_1MB;
 
@@ -2040,6 +2279,608 @@ impl From<NixMFdFlags> for MFdFlags {
     }
 }
 
+//
+// openat2(2) wrapper, nix' does not support Android yet.
+//
+
+bitflags! {
+    /// Flags for openat2(2)
+    #[derive(Copy, Clone, Debug, Eq, Hash, Ord, PartialEq, PartialOrd)]
+    #[repr(transparent)]
+    pub struct ResolveFlag: u64 {
+        /// Do not permit the path resolution to succeed if any component of
+        /// the resolution is not a descendant of the directory indicated by
+        /// dirfd. This causes absolute symbolic links (and absolute values of
+        /// pathname) to be rejected.
+        const RESOLVE_BENEATH = 0x08;
+
+        /// Treat the directory referred to by dirfd as the root directory
+        /// while resolving pathname.
+        const RESOLVE_IN_ROOT = 0x10;
+
+        /// Disallow all magic-link resolution during path resolution. Magic
+        /// links are symbolic link-like objects that are most notably found
+        /// in proc(5); examples include `/proc/[pid]/exe` and `/proc/[pid]/fd/*`.
+        ///
+        /// See symlink(7) for more details.
+        const RESOLVE_NO_MAGICLINKS = 0x02;
+
+        /// Disallow resolution of symbolic links during path resolution. This
+        /// option implies RESOLVE_NO_MAGICLINKS.
+        const RESOLVE_NO_SYMLINKS = 0x04;
+
+        /// Disallow traversal of mount points during path resolution (including
+        /// all bind mounts).
+        const RESOLVE_NO_XDEV = 0x01;
+
+        /// Only complete if resolution can be completed through cached lookup.
+        /// May return EAGAIN if that's not possible.
+        const RESOLVE_CACHED = 0x20;
+    }
+}
+
+/// This structure represents `struct open_how`.
+#[derive(Copy, Clone, Default)]
+#[non_exhaustive]
+#[repr(C)]
+pub struct OpenHow {
+    // open_how flags
+    pub flags: u64,
+    // open_how mode
+    pub mode: u64,
+    // open_how resolve flags
+    pub resolve: u64,
+}
+
+impl OpenHow {
+    /// Create a new zero-filled `open_how`.
+    pub fn new() -> Self {
+        Self::default()
+    }
+
+    /// Set the open flags used to open a file, completely overwriting any
+    /// existing flags.
+    pub fn flags(mut self, flags: OFlag) -> Self {
+        #[expect(clippy::cast_sign_loss)]
+        let flags = flags.bits() as u64;
+        self.flags = flags;
+        self
+    }
+
+    /// Set the file mode new files will be created with, overwriting any
+    /// existing flags.
+    pub fn mode(mut self, mode: Mode) -> Self {
+        let mode = mode.bits().into();
+        self.mode = mode;
+        self
+    }
+
+    /// Set resolve flags, completely overwriting any existing flags.
+    ///
+    /// See [ResolveFlag] for more detail.
+    pub fn resolve(mut self, resolve: ResolveFlag) -> Self {
+        let resolve = resolve.bits();
+        self.resolve = resolve;
+        self
+    }
+}
+
+// Note openat2(2) may not be available,
+// and libc::SYS_openat2 may not be defined.
+// Therefore we query the number using libseccomp.
+static SYS_OPENAT2: LazyLock<Option<libc::c_long>> = LazyLock::new(|| {
+    ScmpSyscall::from_name("openat2")
+        .map(i32::from)
+        .map(libc::c_long::from)
+        .ok()
+});
+
+/// Open or create a file for reading, writing or executing.
+///
+/// `openat2` is an extension of the [`openat`] function that allows the caller
+/// to control how path resolution happens.
+///
+/// # See also
+///
+/// [openat2](https://man7.org/linux/man-pages/man2/openat2.2.html)
+pub fn openat2<P: ?Sized + NixPath, Fd: AsFd>(
+    dirfd: Fd,
+    path: &P,
+    mut how: OpenHow,
+) -> Result<OwnedFd, Errno> {
+    let sys_openat2 = SYS_OPENAT2.ok_or(Errno::ENOSYS)?;
+
+    // SAFETY: In libc we trust.
+    #[expect(clippy::cast_possible_truncation)]
+    let fd = path.with_nix_path(|cstr| unsafe {
+        libc::syscall(
+            sys_openat2,
+            dirfd.as_fd().as_raw_fd(),
+            cstr.as_ptr(),
+            &raw mut how,
+            std::mem::size_of::<OpenHow>(),
+        )
+    })? as RawFd;
+    Errno::result(fd)?;
+
+    // SAFETY:
+    //
+    // `openat2(2)` should return a valid owned fd on success
+    Ok(unsafe { OwnedFd::from_raw_fd(fd) })
+}
+
+bitflags! {
+    /// Mask for defining which events shall be listened with [`Fanotify::mark()`]
+    /// and for querying notifications.
+    #[derive(Copy, Clone, Debug, Eq, Hash, Ord, PartialEq, PartialOrd)]
+    #[repr(transparent)]
+    pub struct MaskFlags: u64 {
+        /// File was accessed
+        const FAN_ACCESS = 0x00000001;
+        /// File was modified
+        const FAN_MODIFY = 0x00000002;
+        /// Metadata changed
+        const FAN_ATTRIB = 0x00000004;
+        /// Writable file closed
+        const FAN_CLOSE_WRITE = 0x00000008;
+        /// Unwritable file closed
+        const FAN_CLOSE_NOWRITE = 0x00000010;
+        /// File was opened
+        const FAN_OPEN = 0x00000020;
+        /// File was moved from X
+        const FAN_MOVED_FROM = 0x00000040;
+        /// File was moved to Y
+        const FAN_MOVED_TO = 0x00000080;
+        /// Subfile was created
+        const FAN_CREATE = 0x00000100;
+        /// Subfile was deleted
+        const FAN_DELETE = 0x00000200;
+        /// Self was deleted
+        const FAN_DELETE_SELF = 0x00000400;
+        /// Self was moved
+        const FAN_MOVE_SELF = 0x00000800;
+        /// File was opened for exec
+        const FAN_OPEN_EXEC = 0x00001000;
+
+        /// Event queued overflowed
+        const FAN_Q_OVERFLOW = 0x00004000;
+        /// Filesystem error
+        const FAN_FS_ERROR = 0x00008000;
+
+        /// File open in perm check
+        const FAN_OPEN_PERM = 0x00010000;
+        /// File accessed in perm check
+        const FAN_ACCESS_PERM = 0x00020000;
+        /// File open/exec in perm check
+        const FAN_OPEN_EXEC_PERM = 0x00040000;
+        // const FAN_DIR_MODIFY = 0x00080000; /* Deprecated (reserved) */
+
+        /// Pre-content access hook
+        const FAN_PRE_ACCESS = 0x00100000;
+        /// Mount was attached
+        const FAN_MNT_ATTACH = 0x01000000;
+        /// Mount was detached
+        const FAN_MNT_DETACH = 0x02000000;
+
+        /// Interested in child events
+        const FAN_EVENT_ON_CHILD = 0x08000000;
+
+        /// File was renamed
+        const FAN_RENAME = 0x10000000;
+
+        /// Event occurred against dir
+        const FAN_ONDIR = 0x40000000;
+
+        /// Close
+        const FAN_CLOSE = Self::FAN_CLOSE_WRITE.bits() | Self::FAN_CLOSE_NOWRITE.bits();
+        /// Moves
+        const FAN_MOVE = Self::FAN_MOVED_FROM.bits() | Self::FAN_MOVED_TO.bits();
+    }
+}
+
+bitflags! {
+    /// Configuration options for [`Fanotify::mark()`].
+    #[derive(Copy, Clone, Debug, Eq, Hash, Ord, PartialEq, PartialOrd)]
+    #[repr(transparent)]
+    pub struct MarkFlags: libc::c_uint {
+        /// Add the events to the marks.
+        const FAN_MARK_ADD = 0x00000001;
+        /// Remove the events to the marks.
+        const FAN_MARK_REMOVE = 0x00000002;
+        /// Don't follow symlinks, mark them.
+        const FAN_MARK_DONT_FOLLOW = 0x00000004;
+        /// Raise an error if filesystem to be marked is not a directory.
+        const FAN_MARK_ONLYDIR = 0x00000008;
+        /// Events added to or removed from the marks.
+        const FAN_MARK_IGNORED_MASK = 0x00000020;
+        /// Ignore mask shall survive modify events.
+        const FAN_MARK_IGNORED_SURV_MODIFY = 0x00000040;
+        /// Remove all marks.
+        const FAN_MARK_FLUSH = 0x00000080;
+        /// Do not pin inode object in the inode cache. Since Linux 5.19.
+        const FAN_MARK_EVICTABLE = 0x00000200;
+        /// Events added to or removed from the marks. Since Linux 6.0.
+        const FAN_MARK_IGNORE = 0x00000400;
+
+        /// Default flag.
+        const FAN_MARK_INODE = 0x00000000;
+        /// Mark the mount specified by pathname.
+        const FAN_MARK_MOUNT = 0x00000010;
+        /// Mark the filesystem specified by pathname. Since Linux 4.20.
+        const FAN_MARK_FILESYSTEM = 0x00000100;
+        /// Mark the mount namespace specified by pathname.
+        const FAN_MARK_MNTNS = 0x00000110;
+
+        /// Combination of `FAN_MARK_IGNORE` and `FAN_MARK_IGNORED_SURV_MODIFY`.
+        const FAN_MARK_IGNORE_SURV = Self::FAN_MARK_IGNORE.bits() | Self::FAN_MARK_IGNORED_SURV_MODIFY.bits();
+    }
+}
+
+/// A fanotify group. This is also a file descriptor that can feed to other
+/// interfaces consuming file descriptors.
+#[derive(Debug)]
+pub struct Fanotify {
+    fd: OwnedFd,
+}
+
+// Note fanotify_mark(2) may not be available,
+// and libc::SYS_fanotify_mark may not be defined.
+// Therefore we query the number using libseccomp.
+static SYS_FANOTIFY_MARK: LazyLock<Option<libc::c_long>> = LazyLock::new(|| {
+    ScmpSyscall::from_name("fanotify_mark")
+        .map(i32::from)
+        .map(libc::c_long::from)
+        .ok()
+});
+
+impl Fanotify {
+    /// Add, remove, or modify an fanotify mark on a filesystem object.
+    ///
+    /// Returns a Result containing either `()` on success or errno otherwise.
+    ///
+    /// For more information, see [fanotify_mark(2)](https://man7.org/linux/man-pages/man7/fanotify_mark.2.html).
+    pub fn mark<Fd: AsFd, P: ?Sized + NixPath>(
+        &self,
+        flags: MarkFlags,
+        mask: MaskFlags,
+        dirfd: Fd,
+        path: Option<&P>,
+    ) -> Result<(), Errno> {
+        let sys_fanotify_mark = SYS_FANOTIFY_MARK.ok_or(Errno::ENOSYS)?;
+
+        // SAFETY: In libc we trust.
+        let res = with_opt_nix_path(path, |p| unsafe {
+            libc::syscall(
+                sys_fanotify_mark,
+                self.fd.as_raw_fd(),
+                flags.bits(),
+                syscall_ll_e!(mask.bits()),
+                dirfd.as_fd().as_raw_fd(),
+                p,
+            )
+        })?;
+
+        Errno::result(res).map(|_| ())
+    }
+}
+
+impl FromRawFd for Fanotify {
+    unsafe fn from_raw_fd(fd: RawFd) -> Self {
+        Fanotify {
+            // SAFETY: This function is unsafe, caller is trusted.
+            fd: unsafe { OwnedFd::from_raw_fd(fd) },
+        }
+    }
+}
+
+impl AsFd for Fanotify {
+    fn as_fd(&'_ self) -> BorrowedFd<'_> {
+        self.fd.as_fd()
+    }
+}
+
+impl AsRawFd for Fanotify {
+    fn as_raw_fd(&self) -> RawFd {
+        self.fd.as_raw_fd()
+    }
+}
+
+impl From<Fanotify> for OwnedFd {
+    fn from(value: Fanotify) -> Self {
+        value.fd
+    }
+}
+
+impl Fanotify {
+    /// Constructs a `Fanotify` wrapping an existing `OwnedFd`.
+    ///
+    /// # Safety
+    ///
+    /// `OwnedFd` is a valid `Fanotify`.
+    pub unsafe fn from_owned_fd(fd: OwnedFd) -> Self {
+        Self { fd }
+    }
+}
+
+pub(crate) fn with_opt_nix_path<P, T, F>(path: Option<&P>, f: F) -> Result<T, Errno>
+where
+    P: ?Sized + NixPath,
+    F: FnOnce(*const libc::c_char) -> T,
+{
+    match path {
+        Some(path) => path.with_nix_path(|p_str| f(p_str.as_ptr())),
+        None => Ok(f(std::ptr::null())),
+    }
+}
+
+bitflags! {
+    /// Mode argument flags for fallocate determining operation performed on a given range.
+    #[derive(Copy, Clone, Debug, Eq, Hash, Ord, PartialEq, PartialOrd)]
+    #[repr(transparent)]
+    pub struct FallocateFlags: libc::c_int {
+        /// File size is not changed.
+        ///
+        /// offset + len can be greater than file size.
+        const FALLOC_FL_KEEP_SIZE = 0x01;
+        /// Deallocates space by creating a hole.
+        ///
+        /// Must be ORed with FALLOC_FL_KEEP_SIZE. Byte range starts at offset and continues for len bytes.
+        const FALLOC_FL_PUNCH_HOLE = 0x02;
+        /// Removes byte range from a file without leaving a hole.
+        ///
+        /// Byte range to collapse starts at offset and continues for len bytes.
+        const FALLOC_FL_COLLAPSE_RANGE = 0x8;
+        /// Zeroes space in specified byte range.
+        ///
+        /// Byte range starts at offset and continues for len bytes.
+        const FALLOC_FL_ZERO_RANGE = 0x10;
+        /// Increases file space by inserting a hole within the file size.
+        ///
+        /// Does not overwrite existing data. Hole starts at offset and continues for len bytes.
+        const FALLOC_FL_INSERT_RANGE = 0x20;
+        /// Shared file data extants are made private to the file.
+        ///
+        /// Guarantees that a subsequent write will not fail due to lack of space.
+        const FALLOC_FL_UNSHARE_RANGE = 0x40;
+        /// Zeroes a specified file range in such a way that subsequent writes to that
+        /// range do not require further changes to the file mapping metadata.
+        ///
+        /// This flag cannot be specified in conjunction with the FALLOC_FL_KEEP_SIZE.
+        const FALLOC_FL_WRITE_ZEROES = 0x80;
+    }
+}
+
+bitflags! {
+    /// Flags used and returned by [`get()`](fn.get.html) and
+    /// [`set()`](fn.set.html).
+    #[derive(Copy, Clone, Debug, Eq, Hash, Ord, PartialEq, PartialOrd)]
+    #[repr(transparent)]
+    pub struct Persona: u64 {
+        /// Provide the legacy virtual address space layout.
+        const ADDR_COMPAT_LAYOUT = 0x0200000;
+        /// Disable address-space-layout randomization.
+        const ADDR_NO_RANDOMIZE = 0x0040000;
+        /// Limit the address space to 32 bits.
+        const ADDR_LIMIT_32BIT = 0x0800000;
+        /// Use `0xc0000000` as the offset at which to search a virtual memory
+        /// chunk on [`mmap(2)`], otherwise use `0xffffe000`.
+        ///
+        /// [`mmap(2)`]: https://man7.org/linux/man-pages/man2/mmap.2.html
+        const ADDR_LIMIT_3GB = 0x8000000;
+        /// User-space function pointers to signal handlers point to descriptors.
+        const FDPIC_FUNCPTRS = 0x0080000;
+        /// Map page 0 as read-only.
+        const MMAP_PAGE_ZERO = 0x0100000;
+        /// `PROT_READ` implies `PROT_EXEC` for [`mmap(2)`].
+        ///
+        /// [`mmap(2)`]: https://man7.org/linux/man-pages/man2/mmap.2.html
+        const READ_IMPLIES_EXEC = 0x0400000;
+        /// No effects.
+        const SHORT_INODE = 0x1000000;
+        /// [`select(2)`], [`pselect(2)`], and [`ppoll(2)`] do not modify the
+        /// returned timeout argument when interrupted by a signal handler.
+        ///
+        /// [`select(2)`]: https://man7.org/linux/man-pages/man2/select.2.html
+        /// [`pselect(2)`]: https://man7.org/linux/man-pages/man2/pselect.2.html
+        /// [`ppoll(2)`]: https://man7.org/linux/man-pages/man2/ppoll.2.html
+        const STICKY_TIMEOUTS = 0x4000000;
+        /// Have [`uname(2)`] report a 2.6.40+ version number rather than a 3.x
+        /// version number.
+        ///
+        /// [`uname(2)`]: https://man7.org/linux/man-pages/man2/uname.2.html
+        const UNAME26 = 0x0020000;
+        /// No effects.
+        const WHOLE_SECONDS = 0x2000000;
+    }
+}
+
+const fn makedev(major: u64, minor: u64) -> libc::dev_t {
+    ((major & 0xffff_f000) << 32)
+        | ((major & 0x0000_0fff) << 8)
+        | ((minor & 0xffff_ff00) << 12)
+        | (minor & 0x0000_00ff)
+}
+
+#[derive(Debug, Clone, Copy, PartialEq, Eq, Hash)]
+#[repr(C)]
+pub(crate) struct seccomp_notif {
+    pub(crate) id: u64,
+    pub(crate) pid: u32,
+    pub(crate) flags: u32,
+    pub(crate) data: seccomp_data,
+}
+
+#[derive(Debug, Clone, Copy, PartialEq, Eq, Hash)]
+#[repr(C)]
+pub(crate) struct seccomp_notif_resp {
+    pub(crate) id: u64,
+    pub(crate) val: i64,
+    pub(crate) error: i32,
+    pub(crate) flags: u32,
+}
+
+#[derive(Debug, Clone, Copy, PartialEq, Eq, Hash)]
+#[repr(C)]
+pub(crate) struct seccomp_notif_addfd {
+    pub(crate) id: u64,
+    pub(crate) flags: u32,
+    pub(crate) srcfd: u32,
+    pub(crate) newfd: u32,
+    pub(crate) newfd_flags: u32,
+}
+
+#[derive(Debug, Clone, Copy, PartialEq, Eq, Hash)]
+#[repr(C)]
+pub(crate) struct seccomp_data {
+    pub(crate) nr: libc::c_int,
+    pub(crate) arch: u32,
+    pub(crate) instruction_pointer: u64,
+    pub(crate) args: [u64; 6],
+}
+
+/// Returns the current parent-death signal.
+pub fn get_pdeathsig() -> Result<Option<Signal>, Errno> {
+    // prctl writes into this var
+    let mut sig: libc::c_int = 0;
+
+    // SAFETY: In libc we trust.
+    let res = unsafe { libc::prctl(libc::PR_GET_PDEATHSIG, &mut sig, 0, 0, 0) };
+
+    match Errno::result(res) {
+        Ok(_) => Ok(match sig {
+            0 => None,
+            _ => Some(Signal::try_from(sig)?),
+        }),
+        Err(e) => Err(e),
+    }
+}
+
+/// Set the parent-death signal of the calling process. This is the signal that the calling process
+/// will get when its parent dies.
+pub fn set_pdeathsig<T: Into<Option<Signal>>>(signal: T) -> Result<(), Errno> {
+    let sig = match signal.into() {
+        Some(s) => s as libc::c_int,
+        None => 0,
+    };
+
+    // SAFETY: In libc we trust.
+    let res = unsafe { libc::prctl(libc::PR_SET_PDEATHSIG, sig, 0, 0, 0) };
+    Errno::result(res).map(drop)
+}
+
+/// Set the dumpable attribute which determines if core dumps are created for this process.
+pub fn set_dumpable(attribute: bool) -> Result<(), Errno> {
+    prctl_set_bool(libc::PR_SET_DUMPABLE, attribute)
+}
+
+/// Set the "child subreaper" attribute for this process.
+pub fn set_child_subreaper(attribute: bool) -> Result<(), Errno> {
+    prctl_set_bool(libc::PR_SET_CHILD_SUBREAPER, attribute)
+}
+
+/// Set the name of the calling thread. Strings longer than 15 bytes will be truncated.
+pub fn set_name(name: &CStr) -> Result<(), Errno> {
+    // SAFETY: In libc we trust.
+    let res = unsafe { libc::prctl(libc::PR_SET_NAME, name.as_ptr(), 0, 0, 0) };
+    Errno::result(res).map(drop)
+}
+
+/// Set the calling threads "no new privs" attribute. Once set this option can not be unset.
+pub fn set_no_new_privs() -> Result<(), Errno> {
+    prctl_set_bool(libc::PR_SET_NO_NEW_PRIVS, true) // Cannot be unset
+}
+
+/// Get the "no new privs" attribute for the calling thread.
+pub fn get_no_new_privs() -> Result<bool, Errno> {
+    prctl_get_bool(libc::PR_GET_NO_NEW_PRIVS)
+}
+
+fn prctl_set_bool(option: libc::c_int, status: bool) -> Result<(), Errno> {
+    // SAFETY: In libc we trust.
+    let res = unsafe { libc::prctl(option, libc::c_ulong::from(status), 0, 0, 0) };
+    Errno::result(res).map(drop)
+}
+
+fn prctl_get_bool(option: libc::c_int) -> Result<bool, Errno> {
+    // SAFETY: In libc we trust.
+    let res = unsafe { libc::prctl(option, 0, 0, 0, 0) };
+    Errno::result(res).map(|res| res != 0)
+}
+
+/// Safe wrapper for dup3(2).
+///
+/// Handles EINTR and returns an OwnedFd.
+pub fn dup3(oldfd: RawFd, newfd: RawFd, flags: libc::c_int) -> Result<OwnedFd, Errno> {
+    // We use SYS_dup3 because Android does not define dup3(2).
+    #[expect(clippy::cast_possible_truncation)]
+    retry_on_eintr(|| {
+        // SAFETY: In libc we trust.
+        Errno::result(unsafe { libc::syscall(libc::SYS_dup3, oldfd, newfd, flags) }).map(|fd| {
+            // SAFETY: dup3(2) returns a valid FD on success.
+            unsafe { OwnedFd::from_raw_fd(fd as RawFd) }
+        })
+    })
+}
+
+/// Set an identifier (or reset it) to the address memory range.
+///
+/// No-op in release mode.
+pub fn set_vma_anon_name(
+    _addr: NonNull<libc::c_void>,
+    _length: NonZeroUsize,
+    _name: Option<&CStr>,
+) -> Result<(), Errno> {
+    #[cfg(not(debug_assertions))]
+    {
+        return Ok(());
+    }
+
+    #[cfg(debug_assertions)]
+    {
+        let nameref = match _name {
+            Some(n) => n.as_ptr(),
+            _ => std::ptr::null(),
+        };
+
+        // SAFETY: In libc we trust.
+        let res = unsafe {
+            libc::prctl(
+                libc::PR_SET_VMA,
+                libc::PR_SET_VMA_ANON_NAME,
+                _addr.as_ptr(),
+                _length,
+                nameref,
+            )
+        };
+        Errno::result(res).map(drop)
+    }
+}
+
+/// An enumeration allowing the definition of the expiration time of an alarm,
+/// recurring or not.
+#[derive(Debug, Clone, Copy, Eq, PartialEq)]
+pub enum Expiration {
+    /// Alarm will trigger once after the time given in `TimeSpec`
+    OneShot(TimeSpec),
+    /// Alarm will trigger after a specified delay and then every interval of
+    /// time.
+    IntervalDelayed(TimeSpec, TimeSpec),
+    /// Alarm will trigger every specified interval of time.
+    Interval(TimeSpec),
+}
+
+bitflags! {
+    /// Flags that are used for arming the timer.
+    #[derive(Copy, Clone, Debug, Eq, Hash, Ord, PartialEq, PartialOrd)]
+    #[repr(transparent)]
+    pub struct TimerSetTimeFlags: libc::c_int {
+        /// Flag TFD_TIMER_ABSTIME.
+        const TFD_TIMER_ABSTIME = libc::TFD_TIMER_ABSTIME;
+        /// Flag TFD_TIMER_CANCEL_ON_SET.
+        const TFD_TIMER_CANCEL_ON_SET = libc::TFD_TIMER_CANCEL_ON_SET;
+    }
+}
+
 /*
  * Constants from <linux/limits.h> not defined by libc yet.
  */
@@ -2050,3 +2891,86 @@ pub(crate) const XATTR_NAME_MAX: usize = 255;
 pub(crate) const XATTR_SIZE_MAX: usize = 1 << 16;
 // size of extended attribute namelist (64k).
 pub(crate) const XATTR_LIST_MAX: usize = 1 << 16;
+
+pub(crate) const MAP_FIXED_NOREPLACE: libc::c_int = 0x100000;
+
+pub(crate) const SHM_EXEC: libc::c_int = 0o100000;
+
+#[cfg(not(any(
+    target_arch = "mips",
+    target_arch = "mips64",
+    target_arch = "mips32r6",
+    target_arch = "mips64r6",
+    target_arch = "sparc"
+)))]
+pub(crate) const TIOCEXCL: Ioctl = 0x540C;
+#[cfg(any(
+    target_arch = "mips",
+    target_arch = "mips64",
+    target_arch = "mips32r6",
+    target_arch = "mips64r6"
+))]
+pub(crate) const TIOCEXCL: Ioctl = 0x740d;
+#[cfg(target_arch = "sparc")]
+pub(crate) const TIOCEXCL: Ioctl = 0x2000740d;
+
+#[cfg(not(any(
+    target_arch = "mips",
+    target_arch = "mips64",
+    target_arch = "mips32r6",
+    target_arch = "mips64r6",
+    target_arch = "sparc"
+)))]
+pub(crate) const TIOCNXCL: Ioctl = 0x540D;
+#[cfg(any(
+    target_arch = "mips",
+    target_arch = "mips64",
+    target_arch = "mips32r6",
+    target_arch = "mips64r6"
+))]
+pub(crate) const TIOCNXCL: Ioctl = 0x740e;
+#[cfg(target_arch = "sparc")]
+pub(crate) const TIOCNXCL: Ioctl = 0x2000740e;
+
+#[cfg(not(any(
+    target_arch = "mips",
+    target_arch = "mips64",
+    target_arch = "mips32r6",
+    target_arch = "mips64r6",
+    target_arch = "powerpc",
+    target_arch = "powerpc64",
+    target_arch = "sparc"
+)))]
+pub(crate) const TIOCGEXCL: Ioctl = 0x80045440;
+#[cfg(any(
+    target_arch = "mips",
+    target_arch = "mips64",
+    target_arch = "mips32r6",
+    target_arch = "mips64r6",
+    target_arch = "powerpc",
+    target_arch = "powerpc64",
+    target_arch = "sparc"
+))]
+pub(crate) const TIOCGEXCL: Ioctl = 0x40045440;
+
+#[cfg(not(any(
+    target_arch = "mips",
+    target_arch = "mips64",
+    target_arch = "mips32r6",
+    target_arch = "mips64r6",
+    target_arch = "powerpc",
+    target_arch = "powerpc64",
+    target_arch = "sparc"
+)))]
+pub(crate) const TIOCGPTPEER: Ioctl = 0x5441;
+#[cfg(any(
+    target_arch = "mips",
+    target_arch = "mips64",
+    target_arch = "mips32r6",
+    target_arch = "mips64r6"
+))]
+pub(crate) const TIOCGPTPEER: Ioctl = 0x20005441;
+#[cfg(any(target_arch = "powerpc", target_arch = "powerpc64"))]
+pub(crate) const TIOCGPTPEER: Ioctl = 0x20005441;
+#[cfg(target_arch = "sparc")]
+pub(crate) const TIOCGPTPEER: Ioctl = 0x20007489;
