@@ -2,22 +2,18 @@
 // Syd: rock-solid application kernel
 // src/kernel/getdents.rs: getdents64(2) handler
 //
-// Copyright (c) 2023, 2024, 2025 Ali Polatel <alip@chesswob.org>
+// Copyright (c) 2023, 2024, 2025, 2026 Ali Polatel <alip@chesswob.org>
 //
 // SPDX-License-Identifier: GPL-3.0
 
-use std::os::fd::RawFd;
+use std::{ops::Deref, os::fd::RawFd};
 
 use libseccomp::ScmpNotifResp;
 use nix::{errno::Errno, NixPath};
 
 use crate::{
-    compat::getdents64,
-    config::{DIRENT_BUF_SIZE, MMAP_MIN_ADDR},
-    kernel::sandbox_path,
-    lookup::CanonicalPath,
-    req::UNotifyEventRequest,
-    sandbox::Capability,
+    compat::getdents64, config::DIRENT_BUF_SIZE, confine::is_valid_ptr, kernel::sandbox_path,
+    lookup::CanonicalPath, req::UNotifyEventRequest, sandbox::Capability,
 };
 
 pub(crate) fn sys_getdents64(request: UNotifyEventRequest) -> ScmpNotifResp {
@@ -38,7 +34,7 @@ pub(crate) fn sys_getdents64(request: UNotifyEventRequest) -> ScmpNotifResp {
         }
 
         // Validate dirp pointer.
-        if req.data.args[1] < *MMAP_MIN_ADDR {
+        if !is_valid_ptr(req.data.args[1], req.data.arch) {
             // SAFETY: If the second argument which must hold a pointer to a
             // linux_dirent structure is not valid, we must return EFAULT
             // without further processing here.
@@ -93,8 +89,7 @@ pub(crate) fn sys_getdents64(request: UNotifyEventRequest) -> ScmpNotifResp {
 
             // Lock sandbox for read to perform Stat access check.
             let sandbox = request.get_sandbox();
-            let safe_name = !sandbox.flags.allow_unsafe_filename();
-            let restrict_mkbdev = !sandbox.flags.allow_unsafe_mkbdev();
+            let check_flags = sandbox.flags.deref().into();
 
             for entry in &mut entries {
                 // SAFETY: Allow the special dot entries `.` and `..`.
@@ -116,8 +111,7 @@ pub(crate) fn sys_getdents64(request: UNotifyEventRequest) -> ScmpNotifResp {
                             pid,
                             Some(&entry.file_type()),
                             Some(entry.as_xpath()),
-                            safe_name,
-                            restrict_mkbdev,
+                            check_flags,
                         )
                         .is_err()
                     {
