@@ -2,20 +2,17 @@
 // Syd: rock-solid application kernel
 // src/kernel/mkdir.rs: mkdir(2) and mkdirat(2) handlers
 //
-// Copyright (c) 2023, 2024, 2025 Ali Polatel <alip@chesswob.org>
+// Copyright (c) 2023, 2024, 2025, 2026 Ali Polatel <alip@chesswob.org>
 //
 // SPDX-License-Identifier: GPL-3.0
 
 use std::os::fd::AsFd;
 
 use libseccomp::ScmpNotifResp;
-use nix::{
-    errno::Errno,
-    sys::stat::{mkdirat, Mode},
-};
+use nix::{errno::Errno, sys::stat::Mode};
 
 use crate::{
-    cookie::safe_umask,
+    cookie::{safe_mkdirat, safe_umask},
     kernel::{syscall_path_handler, to_mode},
     lookup::FsFlags,
     proc::proc_umask,
@@ -89,9 +86,9 @@ fn syscall_mkdir_handler(
     // Note, POSIX ACLs may override this.
     safe_umask(mask);
 
-    mkdirat(
+    safe_mkdirat(
         path.dir.as_ref().map(|fd| fd.as_fd()).ok_or(Errno::EBADF)?,
-        path.base,
+        path.base(),
         mode,
     )
     .map(|_| request.return_syscall(0))
