@@ -2,7 +2,7 @@
 // Syd: rock-solid application kernel
 // src/kernel/ptrace/mmap.rs: ptrace mmap handlers
 //
-// Copyright (c) 2025 Ali Polatel <alip@chesswob.org>
+// Copyright (c) 2025, 2026 Ali Polatel <alip@chesswob.org>
 //
 // SPDX-License-Identifier: GPL-3.0
 
@@ -24,7 +24,7 @@ use crate::{
     config::{PAGE_SIZE, PROC_FILE, PTRACE_DATA_MMAP},
     elf::ExecutableFile,
     error,
-    fs::{fd_status_flags, pidfd_getfd, pidfd_open, PIDFD_THREAD},
+    fd::{fd_status_flags, pidfd_getfd, pidfd_open, PIDFD_THREAD},
     kernel::sandbox_path,
     lookup::{safe_open_msym, CanonicalPath},
     path::XPathBuf,
@@ -45,7 +45,7 @@ pub(crate) fn sysenter_mmap(
     pid: Pid,
     sandbox: &SandboxGuard,
     info: ptrace_syscall_info,
-) -> Result<(), Errno> {
+) -> Result<bool, Errno> {
     let data = if let Some(data) = info.seccomp() {
         data
     } else {
@@ -81,8 +81,8 @@ pub(crate) fn sysenter_mmap(
         && !restrict_append_only
         && (!mem || (mem_max == 0 && mem_vm_max == 0))
     {
-        // Stop at syscall exit.
-        return Ok(());
+        // Continue system call.
+        return Ok(false);
     }
 
     let check_exec = (exec || force || tpe || restrict_exec_memory || restrict_exec_stack)
@@ -325,7 +325,8 @@ pub(crate) fn sysenter_mmap(
     if !mem || (mem_max == 0 && mem_vm_max == 0) {
         // (a) Exec and Memory sandboxing are both disabled.
         // (b) Exec granted access, Memory sandboxing is disabled.
-        return Ok(()); // Stop at syscall exit.
+        // Stop at syscall exit if check_exec is true, otherwise continue.
+        return Ok(check_exec);
     }
 
     // Check VmSize
@@ -378,7 +379,8 @@ pub(crate) fn sysenter_mmap(
             return match mem_act {
                 // Allow cannot happen.
                 Action::Allow => unreachable!(),
-                Action::Warn => Ok(()), // Stop at syscall exit.
+                // Stop at syscall exit if check_exec, otherwise continue.
+                Action::Warn => Ok(check_exec),
                 Action::Panic | Action::Deny | Action::Filter => Err(Errno::ENOMEM),
                 //Do NOT panic the main thread!
                 //Action::Panic => panic!(),
@@ -399,8 +401,8 @@ pub(crate) fn sysenter_mmap(
         }
     }
 
-    // Stop at syscall exit.
-    Ok(())
+    // Stop at syscall exit if check_exec is true, otherwise continue.
+    Ok(check_exec)
 }
 
 #[expect(clippy::cognitive_complexity)]
@@ -409,13 +411,7 @@ pub(crate) fn sysexit_mmap(
     info: ptrace_syscall_info,
     sandbox: &SandboxGuard,
 ) -> Result<(), Errno> {
-    if !sandbox.enabled(Capability::CAP_EXEC) {
-        // Exec sandboxing disabled, nothing to do.
-        // Continue process.
-        return Ok(());
-    }
-
-    // Check for successful sigaction exit.
+    // Check for successful mmap exit.
     match ptrace_get_error(pid, info.arch) {
         Ok(None) => {
             // Successful mmap call, validate proc_pid_maps(5).
