@@ -2,7 +2,7 @@
 // Syd: rock-solid application kernel
 // src/path.rs: Path handling for UNIX
 //
-// Copyright (c) 2024, 2025 Ali Polatel <alip@chesswob.org>
+// Copyright (c) 2024, 2025, 2026 Ali Polatel <alip@chesswob.org>
 // Based in part upon David A. Wheeler's SafeName LSM patches which is:
 //   Copyright (C) 2016 David A. Wheeler
 //   SPDX-License-Identifier: GPL-2.0
@@ -19,10 +19,11 @@ use std::{
         fd::RawFd,
         unix::ffi::{OsStrExt, OsStringExt},
     },
-    path::{Path, PathBuf},
-    sync::LazyLock,
+    path::{Component, Path, PathBuf},
+    sync::{Arc, LazyLock},
 };
 
+use bitflags::bitflags;
 use btoi::btoi;
 use itoa::Integer;
 use memchr::{
@@ -31,15 +32,21 @@ use memchr::{
 };
 use nix::{
     errno::Errno,
-    fcntl::{openat2, OFlag, OpenHow, ResolveFlag, AT_FDCWD},
+    fcntl::{OFlag, AT_FDCWD},
     libc::pid_t,
     unistd::Pid,
     NixPath,
 };
+use tinyvec::TinyVec;
 
 use crate::{
-    config::MAGIC_PREFIX, fs::tgkill, log::log_untrusted_buf, lookup::FileType,
+    compat::{openat2, OpenHow, ResolveFlag},
+    config::MAGIC_PREFIX,
+    fs::tgkill,
+    log::log_untrusted_buf,
+    lookup::FileType,
     retry::retry_on_eintr,
+    sandbox::Flags,
 };
 
 /// Generate a formatted `XPathBuf`.
@@ -56,8 +63,57 @@ pub const PATH_MAX: usize = 4096;
 /// A safe default size to use for paths.
 pub const PATH_MIN: usize = 64;
 
+/// Default capacity of the internal `TinyVec` of `XPathBuf`.
+pub const PATH_CAP: usize = 400;
+
+// This pointer is confined by seccomp for use with execveat(2), fchmodat2(2) and fchownat(2) as empty path.
+// 1. This is an empty NUL-terminated string.
+// 2. Pointer is randomized by ASLR.
+static EMPATH: LazyLock<u64> = LazyLock::new(|| {
+    let path: &'static [u8] = Box::leak(Box::new([0u8]));
+    path.as_ptr() as u64
+});
+
+// This pointer is confined by seccomp for use with execveat(2) as empty argv:
+// 1. Allocate the empty string on heap.
+// 2. Allocate the argv array on heap.
+// 3. Pointers are randomized by ASLR.
+static EMARGV: LazyLock<u64> = LazyLock::new(|| {
+    let empty_str: &'static [u8] = Box::leak(Box::new([0u8]));
+    let empty_ptr = empty_str.as_ptr() as *const libc::c_char;
+    let argv: &'static [*const libc::c_char; 2] =
+        Box::leak(Box::new([empty_ptr, std::ptr::null()]));
+    argv.as_ptr() as u64
+});
+
+// This pointer is confined by seccomp for use with execveat(2) as empty envp.
+// Pointer is randomized by ASLR.
+static EMENVP: LazyLock<u64> = LazyLock::new(|| {
+    let envp: &'static [*const libc::c_char; 1] = Box::leak(Box::new([std::ptr::null()]));
+    envp.as_ptr() as u64
+});
+
 // This pointer is confined by seccomp for use with openat(2) for getdir_long().
-static DOTDOT: LazyLock<u64> = LazyLock::new(|| c"..".as_ptr() as *const libc::c_char as u64);
+// Pointer is randomized by ASLR.
+static DOTDOT: LazyLock<u64> = LazyLock::new(|| {
+    let path: &'static [u8] = Box::leak(Box::new(*b"..\0"));
+    path.as_ptr() as u64
+});
+
+#[inline(always)]
+pub(crate) fn empty_path() -> u64 {
+    *EMPATH
+}
+
+#[inline(always)]
+pub(crate) fn empty_argv() -> u64 {
+    *EMARGV
+}
+
+#[inline(always)]
+pub(crate) fn empty_envp() -> u64 {
+    *EMENVP
+}
 
 #[inline(always)]
 pub(crate) fn dotdot_with_nul() -> u64 {
@@ -68,11 +124,11 @@ pub(crate) fn dotdot_with_nul() -> u64 {
 // SAFETY: k1 == k2 -> hash(k1) == hash(k2) always holds for our PartialEq impl.
 #[allow(clippy::derived_hash_with_manual_eq)]
 #[derive(Clone, Hash, Ord, PartialOrd)]
-pub struct XPathBuf(Vec<u8>);
+pub struct XPathBuf(pub(crate) TinyVec<[u8; PATH_CAP]>);
 
 impl Default for XPathBuf {
     fn default() -> Self {
-        Self(Vec::with_capacity(PATH_MIN))
+        Self(TinyVec::with_capacity(PATH_CAP))
     }
 }
 
@@ -99,14 +155,18 @@ impl PartialEq<XPathBuf> for XPath {
 impl Deref for XPathBuf {
     type Target = XPath;
 
-    #[inline]
     fn deref(&self) -> &XPath {
         XPath::from_bytes(&self.0)
     }
 }
 
 impl Borrow<XPath> for XPathBuf {
-    #[inline]
+    fn borrow(&self) -> &XPath {
+        self.deref()
+    }
+}
+
+impl Borrow<XPath> for Arc<XPathBuf> {
     fn borrow(&self) -> &XPath {
         self.deref()
     }
@@ -173,46 +233,71 @@ impl From<&OsStr> for XPathBuf {
 
 impl From<OsString> for XPathBuf {
     fn from(os: OsString) -> Self {
-        Self(os.into_vec())
+        if os.as_bytes().len() <= PATH_CAP {
+            os.as_bytes().into()
+        } else {
+            Self(TinyVec::Heap(os.into_vec()))
+        }
     }
 }
 
 impl From<String> for XPathBuf {
     fn from(s: String) -> Self {
-        Self(s.as_bytes().into())
+        if s.len() <= PATH_CAP {
+            s.as_bytes().into()
+        } else {
+            Self(TinyVec::Heap(s.into_bytes()))
+        }
     }
 }
 
 impl From<&str> for XPathBuf {
     fn from(s: &str) -> Self {
-        Self(s.as_bytes().into())
+        let mut tv = TinyVec::new();
+        tv.extend_from_slice(s.as_bytes());
+        Self(tv)
     }
 }
 
 impl From<Cow<'_, str>> for XPathBuf {
     fn from(cow: Cow<'_, str>) -> Self {
+        if cow.len() <= PATH_CAP {
+            return cow.as_bytes().into();
+        }
+
         match cow {
-            Cow::Borrowed(s) => Self(s.as_bytes().to_vec()),
-            Cow::Owned(s) => Self(s.into_bytes()),
+            Cow::Borrowed(s) => Self::from(s),
+            Cow::Owned(s) => Self::from(s),
         }
     }
 }
 
 impl From<&[u8]> for XPathBuf {
     fn from(bytes: &[u8]) -> Self {
-        Self(bytes.to_vec())
+        let mut tv = TinyVec::new();
+        tv.extend_from_slice(bytes);
+        Self(tv)
     }
 }
 
 impl From<Vec<u8>> for XPathBuf {
     fn from(vec: Vec<u8>) -> Self {
-        Self(vec)
+        if vec.len() <= PATH_CAP {
+            vec.as_slice().into()
+        } else {
+            Self(TinyVec::Heap(vec))
+        }
     }
 }
 
 impl From<VecDeque<u8>> for XPathBuf {
-    fn from(vec: VecDeque<u8>) -> Self {
-        Self(vec.into())
+    fn from(mut vec: VecDeque<u8>) -> Self {
+        if vec.len() <= PATH_CAP {
+            vec.make_contiguous();
+            vec.as_slices().0.into()
+        } else {
+            Self(TinyVec::Heap(Vec::from(vec)))
+        }
     }
 }
 
@@ -325,6 +410,58 @@ impl NixPath for XPath {
     }
 }
 
+bitflags! {
+    /// Flags for XPath::check().
+    #[derive(Clone, Copy, Debug, Default, Eq, PartialEq)]
+    pub struct XPathCheckFlags: u8 {
+        /// Restrict file names to allowed characters.
+        const SAFE_NAME = 1 << 0;
+        /// Restrict listing block devices.
+        const RESTRICT_MKBDEV = 1 << 1;
+        /// Restrict access to proc(5) magic links.
+        const RESTRICT_MAGICLINKS = 1 << 2;
+    }
+}
+
+impl XPathCheckFlags {
+    /// Returns true if safe name check is enabled.
+    pub fn safe_name(self) -> bool {
+        self.contains(Self::SAFE_NAME)
+    }
+
+    /// Returns true if block device restriction is enabled.
+    pub fn restrict_mkbdev(self) -> bool {
+        self.contains(Self::RESTRICT_MKBDEV)
+    }
+
+    /// Returns true if magic link restriction is enabled.
+    pub fn restrict_magiclinks(self) -> bool {
+        self.contains(Self::RESTRICT_MAGICLINKS)
+    }
+}
+
+impl From<&Flags> for XPathCheckFlags {
+    fn from(flags: &Flags) -> Self {
+        let mut check = Self::empty();
+        if !flags.allow_unsafe_filename() {
+            check.insert(Self::SAFE_NAME);
+        }
+        if !flags.allow_unsafe_mkbdev() {
+            check.insert(Self::RESTRICT_MKBDEV);
+        }
+        if flags.force_no_magiclinks() {
+            check.insert(Self::RESTRICT_MAGICLINKS);
+        }
+        check
+    }
+}
+
+impl From<Flags> for XPathCheckFlags {
+    fn from(flags: Flags) -> Self {
+        (&flags).into()
+    }
+}
+
 impl XPath {
     /// Detects unsafe paths.
     ///
@@ -340,14 +477,12 @@ impl XPath {
     ///
     /// Note, returning error here denies access,
     /// regardless of the state of sandboxing.
-    #[inline(always)]
     pub fn check(
         &self,
         pid: Pid,
         file_type: Option<&FileType>,
         dir_entry: Option<&XPath>,
-        safe_name: bool,
-        restrict_mkbdev: bool,
+        flags: XPathCheckFlags,
     ) -> Result<(), Errno> {
         //
         // RESTRICTION 1: Prevent listing block devices and files of unknown type.
@@ -355,7 +490,7 @@ impl XPath {
         // SAFETY: Prevent listing block devices and files of unknown type.
         // Block device restriction may be relaxed with trace/allow_unsafe_mkbdev:1.
         if file_type == Some(&FileType::Unk)
-            || (restrict_mkbdev && file_type == Some(&FileType::Blk))
+            || (flags.restrict_mkbdev() && file_type == Some(&FileType::Blk))
         {
             return Err(Errno::ENOENT);
         }
@@ -368,10 +503,11 @@ impl XPath {
         // Note, we skip checking procfs so we don't prevent pipe/socket access
         // unintentionally. Similarly we skip checking memory fd names which do not
         // represent actual file paths.
+        // As of version 3.48.0, we return EILSEQ rather than EINVAL which matches ZFS.
         let is_mfd = matches!(file_type, Some(FileType::Mfd));
         let is_proc_dir = self.starts_with(b"/proc");
-        if safe_name && !is_mfd && !is_proc_dir && self.check_name().is_err() {
-            return Err(Errno::EINVAL);
+        if flags.safe_name() && !is_mfd && !is_proc_dir {
+            self.check_name()?;
         }
         // END OF RESTRICTION 2
 
@@ -404,6 +540,21 @@ impl XPath {
         } else {
             return Ok(());
         };
+
+        // RESTRICTION 3: Prevent access to magiclinks with trace/force_no_magiclinks:1.
+        //
+        // Exception: /proc/self and /proc/thread-self aren't considered magiclinks.
+        // This is consistent with the behaviour of the RESOLVE_NO_MAGICLINKS openat2(2) resolve flag.
+        if flags.restrict_magiclinks()
+            && matches!(file_type, Some(t) if t.is_link())
+            && !self.is_proc_self(false)
+            && !self.is_proc_self(true)
+        {
+            return Err(Errno::ENOENT);
+        }
+        // END OF RESTRICTION 3
+
+        // Remaining restrictions apply to /proc/$pid where $pid != caller only.
         let proc_pid = if let Some(pid) = proc_pid {
             Pid::from_raw(pid)
         } else {
@@ -414,16 +565,16 @@ impl XPath {
         }
 
         //
-        // RESTRICTION 2: Protect readdir(/proc).
+        // RESTRICTION 4: Protect readdir(/proc) from enumerating PIDs.
         //
         // SAFETY: Prevent /proc process tree traversal.
         if is_proc {
             return Err(Errno::ENOENT);
         }
-        // END OF RESTRICTION 2
+        // END OF RESTRICTION 4
 
         //
-        // RESTRICTION 3: Protect Syd procfs.
+        // RESTRICTION 5: Protect Syd procfs.
         //
         // SAFETY: Protect Syd /proc directory!
         //
@@ -437,7 +588,7 @@ impl XPath {
         if tgkill(syd_pid, proc_pid, 0).is_ok() {
             return Err(Errno::ENOENT);
         }
-        // END OF RESTRICTION 3
+        // END OF RESTRICTION 5
 
         // TODO: Add more restrictions as needed.
         Ok(())
@@ -475,11 +626,11 @@ impl XPath {
     /// # Returns
     ///
     /// * `Ok(())` if the filename is valid and safe.
-    /// * `Err(Errno::EINVAL)` if the filename is invalid or unsafe.
+    /// * `Err(Errno::EILSEQ)` if the filename is invalid or unsafe.
     ///
     /// # Errors
     ///
-    /// Returns `Err(Errno::EINVAL)` if any of the validation rules are not met.
+    /// Returns `Err(Errno::EILSEQ)` if any of the validation rules are not met.
     ///
     /// # Security
     ///
@@ -493,11 +644,11 @@ impl XPath {
         let len = name.len();
 
         if len == 0 {
-            return Err(Errno::EINVAL);
+            return Err(Errno::EILSEQ);
         }
 
         // Check if the filename is valid UTF-8.
-        let name_utf8 = std::str::from_utf8(name).or(Err(Errno::EINVAL))?;
+        let name_utf8 = std::str::from_utf8(name).or(Err(Errno::EILSEQ))?;
 
         // Check if first and last character is not whitespace.
         // This includes UTF-8 whitespace.
@@ -507,7 +658,7 @@ impl XPath {
             .map(|c| c.is_whitespace())
             .unwrap_or(false)
         {
-            return Err(Errno::EINVAL);
+            return Err(Errno::EILSEQ);
         }
         if name_utf8
             .chars()
@@ -515,7 +666,7 @@ impl XPath {
             .map(|c| c.is_whitespace())
             .unwrap_or(false)
         {
-            return Err(Errno::EINVAL);
+            return Err(Errno::EILSEQ);
         }
 
         let first_byte = name[0];
@@ -523,7 +674,7 @@ impl XPath {
 
         // Check the first byte.
         if !is_permitted_initial(first_byte) {
-            return Err(Errno::EINVAL);
+            return Err(Errno::EILSEQ);
         }
 
         // Check the middle bytes (if any).
@@ -532,13 +683,13 @@ impl XPath {
                 // Only one middle byte to check.
                 let middle_byte = name[1];
                 if !is_permitted_middle(middle_byte) {
-                    return Err(Errno::EINVAL);
+                    return Err(Errno::EILSEQ);
                 }
             }
             n if n > 2 => {
                 for &b in &name[1..len - 1] {
                     if !is_permitted_middle(b) {
-                        return Err(Errno::EINVAL);
+                        return Err(Errno::EILSEQ);
                     }
                 }
             }
@@ -547,12 +698,41 @@ impl XPath {
 
         // Check the last byte.
         if !is_permitted_final(last_byte) {
-            return Err(Errno::EINVAL);
+            return Err(Errno::EILSEQ);
         }
 
         Ok(())
     }
 
+    /// Convert /proc/${pid} to /proc/self as necessary.
+    ///
+    /// This must never used in syscalls only in access check.
+    pub fn replace_proc_self<'a>(&'a self, pid: Pid) -> Cow<'a, Self> {
+        let p = if let Some(p) = self.split_prefix(b"/proc") {
+            p
+        } else {
+            return Cow::Borrowed(self);
+        };
+
+        let mut buf = itoa::Buffer::new();
+        let pid = buf.format(pid.as_raw());
+        let p = if let Some(p) = p.split_prefix(pid.as_bytes()) {
+            p
+        } else {
+            return Cow::Borrowed(self);
+        };
+
+        let mut pdir = XPathBuf::from("/proc/self");
+        // SAFETY: We do not use XPathBuf::push here,
+        // and allow slash in path. This is safe against
+        // path traversals because the path is guaranteed
+        // to be a canonicalized path.
+        pdir.append_byte(b'/');
+        pdir.append_bytes(p.as_bytes());
+
+        Cow::Owned(pdir)
+    }
+
     /// Returns a path that, when joined onto `base`, yields `self`.
     ///
     /// Expects normalized, canonical path.
@@ -717,6 +897,33 @@ impl XPath {
         }
     }
 
+    /// Returns a relative path by stripping the root component.
+    ///
+    /// # Safety
+    ///
+    /// This function does NOT assume `self` is normalized and performs
+    /// a small amount of normalization. Therefore this function is safe
+    /// to use with non-normalized paths.
+    pub fn strip_root(&self) -> XPathBuf {
+        let src = self.as_path();
+        let mut dst = XPathBuf::new();
+
+        for comp in src.components() {
+            if matches!(
+                comp,
+                Component::RootDir | Component::Prefix(_) | Component::CurDir
+            ) {
+                continue;
+            }
+            if !dst.is_empty() {
+                dst.append_byte(b'/');
+            }
+            dst.append_bytes(comp.as_os_str().as_bytes());
+        }
+
+        dst
+    }
+
     /// Returns a path that, when joined onto `base`, yields `self`.
     ///
     /// # Safety
@@ -726,7 +933,7 @@ impl XPath {
     /// # Errors
     ///
     /// If `base` is not a prefix of self (i.e., `starts_with` returns
-    /// `false`), returns `Err`.
+    /// `false`), returns `None`.
     pub fn strip_prefix(&self, base: &[u8]) -> Option<&Self> {
         if !self.starts_with(base) {
             return None;
@@ -815,6 +1022,39 @@ impl XPath {
         false
     }
 
+    /// Check if path is a valid Syd glob.
+    pub fn is_glob(&self) -> bool {
+        match self.first() {
+            Some(b'/') => true, // Absolute path.
+            Some(b'@') => true, // Abstract socket path.
+            // We reserve the prefix `!' for some special features:
+            //
+            // 1. Sending file descriptors to unnamed sockets may be
+            //    allowed with `allow/net/sendfd+!unnamed`.
+            // 2. Binding/Connecting to unnamed UNIX sockets may be
+            //    allowed with e.g. `allow/net/bind+!unnamed`.
+            // 3. Confining memory file descriptors with
+            //    `allow/create+!memfd:*` or
+            //    `allow/truncate+!memfd-hugetlb:*`.
+            Some(b'!') => self.is_special(),
+            _ => {
+                // We match on canonicalized paths,
+                // relative patterns are a common case of error,
+                // let's just prevent them until someone comes
+                // up with a valid usecase.
+                false
+            }
+        }
+    }
+
+    /// Return true if the pattern is a reserved Syd path.
+    #[inline]
+    pub fn is_special(&self) -> bool {
+        self.is_equal(b"!unnamed")
+            || self.starts_with(b"!memfd:")
+            || self.starts_with(b"!memfd-hugetlb:")
+    }
+
     /// Check if path starts with the `MAGIC_PREFIX`.
     pub fn is_magic(&self) -> bool {
         self.starts_with(MAGIC_PREFIX)
@@ -846,6 +1086,27 @@ impl XPath {
         }
     }
 
+    /// Check if path points to a mountpoint inside proc(5).
+    pub fn is_proc_mnt(&self) -> bool {
+        // /proc/sys/fs/binfmt_misc
+        if self.is_equal(b"binfmt_misc") {
+            return true;
+        }
+        if self.ends_with(b"/binfmt_misc") {
+            return true;
+        }
+
+        // /proc/fs/nfsd
+        if self.is_equal(b"nfsd") {
+            return true;
+        }
+        if self.ends_with(b"/nfsd") {
+            return true;
+        }
+
+        false
+    }
+
     /// Check if path points to devfs, ie. starts with `/dev`.
     /// The literal path `/dev` returns false.
     ///
@@ -891,7 +1152,13 @@ impl XPath {
 
     /// Check if path points to machine-id(5), /etc/hostid or /var/adm/hostid.
     pub fn is_machine_id(&self) -> bool {
-        const MACHINE_ID: &[&[u8]] = &[b"/etc/machine-id", b"/etc/hostid", b"/var/adm/hostid"];
+        const MACHINE_ID: &[&[u8]] = &[
+            b"/etc/machine-id",
+            b"/etc/hostid",
+            b"/var/adm/hostid",
+            b"/sys/class/dmi/id/product_uuid",
+            b"/sys/devices/virtual/dmi/id/product_uuid",
+        ];
         MACHINE_ID.iter().any(|f| self.is_equal(f))
     }
 
@@ -1061,7 +1328,6 @@ impl XPath {
     ///
     /// Use the bytes up to the first nul byte if one exists;
     /// otherwise use the whole byte slice.
-    #[inline]
     pub fn from_bytes_until_nul(slice: &[u8]) -> &XPath {
         // Find first nul byte if any.
         // Use the whole slice otherwise.
@@ -1385,7 +1651,6 @@ impl XPathBuf {
     ///
     /// 1. Path must be a normalized absolute path!
     /// 2. Path must not have a trailing slash!
-    #[inline]
     pub unsafe fn pop_unchecked(&mut self) {
         #[expect(clippy::arithmetic_side_effects)]
         if let Some(idx) = memrchr(b'/', &self.as_bytes()[1..]) {
@@ -1410,14 +1675,19 @@ impl XPathBuf {
         self.0.pop()
     }
 
+    /// Clear the internal `Vec`.
+    pub fn clear(&mut self) {
+        self.0.clear()
+    }
+
     /// Convert a `XPathBuf` to a `Vec`.
     pub fn into_vec(self) -> Vec<u8> {
-        self.0
+        self.0.to_vec()
     }
 
     /// Convert a `XPathBuf` to an `OsString`.
     pub fn into_os_string(self) -> OsString {
-        OsString::from_vec(self.0)
+        OsString::from_vec(self.0.to_vec())
     }
 
     /// Shorten the vector, keeping the first len elements and dropping
@@ -1508,25 +1778,21 @@ impl XPathBuf {
     }
 
     /// Returns a slice containing the entire path buffer.
-    #[inline]
     pub fn as_slice(&self) -> &[u8] {
         self.0.as_slice()
     }
 
     /// Returns a mutable slice containing the entire path buffer.
-    #[inline]
     pub fn as_mut_slice(&mut self) -> &mut [u8] {
         self.0.as_mut_slice()
     }
 
     /// Returns a pointer to the internal `Vec`.
-    #[inline]
     pub fn as_ptr(&self) -> *const u8 {
         self.0.as_ptr()
     }
 
     /// Returns a mutable pointer to the internal `Vec`.
-    #[inline]
     pub fn as_mut_ptr(&mut self) -> *mut u8 {
         self.0.as_mut_ptr()
     }
@@ -1537,20 +1803,9 @@ impl XPathBuf {
         self.0[idx] = val;
     }
 
-    /// Forces the length of the internal `Vec`.
-    ///
-    /// # Safety
-    ///
-    /// - `new_len` must be less than or equal to `capacity()`.
-    /// - The elements at `old_len..new_len` must be initialized.
-    #[inline]
-    pub unsafe fn set_len(&mut self, new_len: usize) {
-        self.0.set_len(new_len)
-    }
-
     /// Construct an empty `XPathBuf` with capacity pre-allocated.
     pub fn with_capacity(n: usize) -> Self {
-        Self(Vec::with_capacity(n))
+        Self(TinyVec::with_capacity(n))
     }
 
     /// Report capacity of path.
@@ -1561,42 +1816,36 @@ impl XPathBuf {
     /// Constructs a new, empty `XPathBuf`.
     ///
     /// The internal vector will not allocate until elements are pushed onto it.
-    pub const fn empty() -> Self {
+    pub fn empty() -> Self {
         Self::new()
     }
 
     /// Constructs a new, empty `XPathBuf`.
     ///
     /// The internal vector will not allocate until elements are pushed onto it.
-    pub const fn new() -> XPathBuf {
-        Self(Vec::new())
+    pub fn new() -> XPathBuf {
+        Self(TinyVec::new())
     }
 }
 
 /// Logs an untrusted Path, escaping it as hex if it contains control
 /// characters.
-#[inline]
-pub fn mask_path(path: &Path) -> String {
-    let (mask, _) = log_untrusted_buf(path.as_os_str().as_bytes());
-    mask
+pub fn mask_path<P: AsRef<Path> + ?Sized>(path: &P) -> String {
+    log_untrusted_buf(path.as_ref().as_os_str().as_bytes()).0
 }
 
-#[inline]
 fn is_permitted_initial(b: u8) -> bool {
     is_permitted_byte(b) && !matches!(b, b'-' | b' ' | b'~')
 }
 
-#[inline]
 fn is_permitted_middle(b: u8) -> bool {
     is_permitted_byte(b)
 }
 
-#[inline]
 fn is_permitted_final(b: u8) -> bool {
     is_permitted_byte(b) && b != b' '
 }
 
-#[inline]
 fn is_permitted_byte(b: u8) -> bool {
     // As of version 3.38.0 the following characters are removed:
     // 1. `:` which is used commonly across devfs and procfs.
@@ -1617,6 +1866,7 @@ mod tests {
     use nix::unistd::{gettid, pause};
 
     use super::*;
+    use crate::magic::ProcMagic;
 
     struct CCSTestCase<'a> {
         src: &'a str,
@@ -1865,18 +2115,88 @@ mod tests {
         }
     }
 
+    #[test]
+    fn test_strip_root() {
+        let cases = [
+            // Basic relative & empty
+            ("", ""),
+            (".", ""),
+            ("./", ""),
+            ("./.", ""),
+            ("././", ""),
+            ("foo", "foo"),
+            ("foo/bar", "foo/bar"),
+            ("foo/./bar", "foo/bar"),
+            ("./foo/bar", "foo/bar"),
+            // Leading root & redundant separators / dots
+            ("/", ""),
+            ("//", ""),
+            ("///", ""),
+            ("/.", ""),
+            ("/./", ""),
+            ("/./.", ""),
+            ("/././", ""),
+            ("/foo", "foo"),
+            ("/foo/", "foo"),
+            ("//foo", "foo"),
+            ("///foo", "foo"),
+            ("/./foo", "foo"),
+            ("/././foo", "foo"),
+            ("/foo/bar", "foo/bar"),
+            ("/foo/bar/", "foo/bar"),
+            ("//foo//bar///", "foo/bar"),
+            ("/./foo/./bar/.", "foo/bar"),
+            ("///.//././foo/bar", "foo/bar"),
+            // Dots inside segment names must be preserved
+            ("/foo.bar", "foo.bar"),
+            ("/foo..bar", "foo..bar"),
+            // Parent directory segments must be preserved (no over-normalization)
+            ("../foo", "../foo"),
+            ("../../foo", "../../foo"),
+            ("foo/../bar", "foo/../bar"),
+            ("/..", ".."),
+            ("/../", ".."),
+            ("/../foo", "../foo"),
+            ("/foo/..", "foo/.."),
+            ("/foo/../bar", "foo/../bar"),
+            ("/foo/../../bar", "foo/../../bar"),
+            ("/../../etc/passwd", "../../etc/passwd"),
+            ("/foo/../../../etc/passwd", "foo/../../../etc/passwd"),
+            // "Weird" but valid segments
+            ("/ foo", " foo"),
+            ("/\u{00A0}foo", "\u{00A0}foo"), // non-breaking space
+        ];
+
+        for &(input, expected) in &cases {
+            let path = XPath::from_bytes(input.as_bytes());
+            let stripped = path.strip_root();
+            let expected_path = XPath::from_bytes(expected.as_bytes());
+
+            assert_eq!(
+                stripped.as_xpath(),
+                expected_path,
+                "strip_root failed: input={input:?}, expected={expected:?}, got={:?}",
+                stripped
+            );
+        }
+    }
+
     #[test]
     fn test_path_check_file_type() {
         xpath!("/proc")
-            .check(Pid::from_raw(1), Some(&FileType::Dir), None, true, true)
+            .check(
+                Pid::from_raw(1),
+                Some(&FileType::Dir),
+                None,
+                XPathCheckFlags::SAFE_NAME | XPathCheckFlags::RESTRICT_MKBDEV,
+            )
             .unwrap();
         xpath!("/proc")
             .check(
                 Pid::from_raw(1),
                 Some(&FileType::Dir),
                 Some(&XPath::from_bytes(b"self")),
-                true,
-                true,
+                XPathCheckFlags::SAFE_NAME | XPathCheckFlags::RESTRICT_MKBDEV,
             )
             .unwrap();
         xpath!("/proc")
@@ -1884,24 +2204,48 @@ mod tests {
                 Pid::from_raw(1),
                 Some(&FileType::Reg),
                 Some(&XPath::from_bytes(b"uptime")),
-                true,
-                true,
+                XPathCheckFlags::SAFE_NAME | XPathCheckFlags::RESTRICT_MKBDEV,
             )
             .unwrap();
         xpath!("/dev/null")
-            .check(Pid::from_raw(1), Some(&FileType::Chr), None, true, true)
+            .check(
+                Pid::from_raw(1),
+                Some(&FileType::Chr),
+                None,
+                XPathCheckFlags::SAFE_NAME | XPathCheckFlags::RESTRICT_MKBDEV,
+            )
             .unwrap();
         xpath!("/dev/log")
-            .check(Pid::from_raw(1), Some(&FileType::Sock), None, true, true)
+            .check(
+                Pid::from_raw(1),
+                Some(&FileType::Sock),
+                None,
+                XPathCheckFlags::SAFE_NAME | XPathCheckFlags::RESTRICT_MKBDEV,
+            )
             .unwrap();
         xpath!("/dev/fifo")
-            .check(Pid::from_raw(1), Some(&FileType::Fifo), None, true, true)
+            .check(
+                Pid::from_raw(1),
+                Some(&FileType::Fifo),
+                None,
+                XPathCheckFlags::SAFE_NAME | XPathCheckFlags::RESTRICT_MKBDEV,
+            )
             .unwrap();
         xpath!("/dev/sda1")
-            .check(Pid::from_raw(1), Some(&FileType::Blk), None, true, true)
+            .check(
+                Pid::from_raw(1),
+                Some(&FileType::Blk),
+                None,
+                XPathCheckFlags::SAFE_NAME | XPathCheckFlags::RESTRICT_MKBDEV,
+            )
             .unwrap_err();
         xpath!("/dev/lmao")
-            .check(Pid::from_raw(1), Some(&FileType::Unk), None, true, true)
+            .check(
+                Pid::from_raw(1),
+                Some(&FileType::Unk),
+                None,
+                XPathCheckFlags::SAFE_NAME | XPathCheckFlags::RESTRICT_MKBDEV,
+            )
             .unwrap_err();
     }
 
@@ -1914,8 +2258,7 @@ mod tests {
                 this,
                 Some(&FileType::Dir),
                 Some(&xpath!("{this}")),
-                true,
-                true,
+                XPathCheckFlags::SAFE_NAME | XPathCheckFlags::RESTRICT_MKBDEV,
             )
             .unwrap();
         xpath!("/proc/{this}")
@@ -1923,8 +2266,7 @@ mod tests {
                 this,
                 Some(&FileType::Reg),
                 Some(&XPath::from_bytes(b"mem")),
-                true,
-                true,
+                XPathCheckFlags::SAFE_NAME | XPathCheckFlags::RESTRICT_MKBDEV,
             )
             .unwrap();
         xpath!("/proc/{this}")
@@ -1932,8 +2274,7 @@ mod tests {
                 this,
                 Some(&FileType::Dir),
                 Some(&XPath::from_bytes(b"")),
-                true,
-                true,
+                XPathCheckFlags::SAFE_NAME | XPathCheckFlags::RESTRICT_MKBDEV,
             )
             .unwrap();
         xpath!("/proc/{this}/task")
@@ -1941,8 +2282,7 @@ mod tests {
                 this,
                 Some(&FileType::Dir),
                 Some(&xpath!("{this}")),
-                true,
-                true,
+                XPathCheckFlags::SAFE_NAME | XPathCheckFlags::RESTRICT_MKBDEV,
             )
             .unwrap();
         xpath!("/proc")
@@ -1950,8 +2290,7 @@ mod tests {
                 this,
                 Some(&FileType::Dir),
                 Some(&xpath!("{that}")),
-                true,
-                true,
+                XPathCheckFlags::SAFE_NAME | XPathCheckFlags::RESTRICT_MKBDEV,
             )
             .unwrap_err();
         xpath!("/proc/{that}")
@@ -1959,8 +2298,7 @@ mod tests {
                 this,
                 Some(&FileType::Reg),
                 Some(&XPath::from_bytes(b"")),
-                true,
-                true,
+                XPathCheckFlags::SAFE_NAME | XPathCheckFlags::RESTRICT_MKBDEV,
             )
             .unwrap();
         xpath!("/proc/{that}")
@@ -1968,8 +2306,7 @@ mod tests {
                 this,
                 Some(&FileType::Dir),
                 Some(&XPath::from_bytes(b"")),
-                true,
-                true,
+                XPathCheckFlags::SAFE_NAME | XPathCheckFlags::RESTRICT_MKBDEV,
             )
             .unwrap();
         xpath!("/proc/{that}/task")
@@ -1977,8 +2314,7 @@ mod tests {
                 this,
                 Some(&FileType::Dir),
                 Some(&xpath!("{that}")),
-                true,
-                true,
+                XPathCheckFlags::SAFE_NAME | XPathCheckFlags::RESTRICT_MKBDEV,
             )
             .unwrap();
     }
@@ -1992,8 +2328,7 @@ mod tests {
                 syd,
                 Some(&FileType::Dir),
                 Some(&xpath!("{syd}")),
-                true,
-                true,
+                XPathCheckFlags::SAFE_NAME | XPathCheckFlags::RESTRICT_MKBDEV,
             )
             .unwrap();
         xpath!("/proc")
@@ -2001,8 +2336,7 @@ mod tests {
                 pid,
                 Some(&FileType::Dir),
                 Some(&xpath!("{syd}")),
-                true,
-                true,
+                XPathCheckFlags::SAFE_NAME | XPathCheckFlags::RESTRICT_MKBDEV,
             )
             .unwrap_err();
         xpath!("/proc/{syd}")
@@ -2010,8 +2344,7 @@ mod tests {
                 syd,
                 Some(&FileType::Reg),
                 Some(&XPath::from_bytes(b"")),
-                true,
-                true,
+                XPathCheckFlags::SAFE_NAME | XPathCheckFlags::RESTRICT_MKBDEV,
             )
             .unwrap();
         xpath!("/proc/{syd}")
@@ -2019,8 +2352,7 @@ mod tests {
                 pid,
                 Some(&FileType::Reg),
                 Some(&XPath::from_bytes(b"")),
-                true,
-                true,
+                XPathCheckFlags::SAFE_NAME | XPathCheckFlags::RESTRICT_MKBDEV,
             )
             .unwrap_err();
         xpath!("/proc/{syd}")
@@ -2028,8 +2360,7 @@ mod tests {
                 pid,
                 Some(&FileType::Dir),
                 Some(&XPath::from_bytes(b"")),
-                true,
-                true,
+                XPathCheckFlags::SAFE_NAME | XPathCheckFlags::RESTRICT_MKBDEV,
             )
             .unwrap_err();
         xpath!("/proc/{syd}/task")
@@ -2037,8 +2368,7 @@ mod tests {
                 syd,
                 Some(&FileType::Dir),
                 Some(&xpath!("{syd}")),
-                true,
-                true,
+                XPathCheckFlags::SAFE_NAME | XPathCheckFlags::RESTRICT_MKBDEV,
             )
             .unwrap();
         xpath!("/proc/{syd}/task")
@@ -2046,8 +2376,7 @@ mod tests {
                 pid,
                 Some(&FileType::Dir),
                 Some(&xpath!("{syd}")),
-                true,
-                true,
+                XPathCheckFlags::SAFE_NAME | XPathCheckFlags::RESTRICT_MKBDEV,
             )
             .unwrap_err();
     }
@@ -2069,8 +2398,7 @@ mod tests {
                 tid,
                 Some(&FileType::Dir),
                 Some(&xpath!("{tid}")),
-                true,
-                true,
+                XPathCheckFlags::SAFE_NAME | XPathCheckFlags::RESTRICT_MKBDEV,
             )
             .unwrap();
         xpath!("/proc/{tid}")
@@ -2078,8 +2406,7 @@ mod tests {
                 tid,
                 Some(&FileType::Reg),
                 Some(&XPath::from_bytes(b"")),
-                true,
-                true,
+                XPathCheckFlags::SAFE_NAME | XPathCheckFlags::RESTRICT_MKBDEV,
             )
             .unwrap();
         xpath!("/proc/{tid}")
@@ -2087,8 +2414,7 @@ mod tests {
                 tid,
                 Some(&FileType::Dir),
                 Some(&XPath::from_bytes(b"")),
-                true,
-                true,
+                XPathCheckFlags::SAFE_NAME | XPathCheckFlags::RESTRICT_MKBDEV,
             )
             .unwrap();
         xpath!("/proc/{tid}/task")
@@ -2096,8 +2422,7 @@ mod tests {
                 tid,
                 Some(&FileType::Dir),
                 Some(&xpath!("{tid}")),
-                true,
-                true,
+                XPathCheckFlags::SAFE_NAME | XPathCheckFlags::RESTRICT_MKBDEV,
             )
             .unwrap();
         xpath!("/proc")
@@ -2105,8 +2430,7 @@ mod tests {
                 pid,
                 Some(&FileType::Dir),
                 Some(&xpath!("{tid}")),
-                true,
-                true,
+                XPathCheckFlags::SAFE_NAME | XPathCheckFlags::RESTRICT_MKBDEV,
             )
             .unwrap_err();
         xpath!("/proc/{tid}")
@@ -2114,8 +2438,7 @@ mod tests {
                 pid,
                 Some(&FileType::Reg),
                 Some(&XPath::from_bytes(b"")),
-                true,
-                true,
+                XPathCheckFlags::SAFE_NAME | XPathCheckFlags::RESTRICT_MKBDEV,
             )
             .unwrap_err();
         xpath!("/proc/{tid}")
@@ -2123,8 +2446,7 @@ mod tests {
                 pid,
                 Some(&FileType::Dir),
                 Some(&XPath::from_bytes(b"")),
-                true,
-                true,
+                XPathCheckFlags::SAFE_NAME | XPathCheckFlags::RESTRICT_MKBDEV,
             )
             .unwrap_err();
         xpath!("/proc/{tid}/task")
@@ -2132,12 +2454,75 @@ mod tests {
                 pid,
                 Some(&FileType::Dir),
                 Some(&xpath!("{tid}")),
-                true,
-                true,
+                XPathCheckFlags::SAFE_NAME | XPathCheckFlags::RESTRICT_MKBDEV,
             )
             .unwrap_err();
     }
 
+    #[test]
+    fn test_path_check_restrict_magiclinks() {
+        let pid = Pid::from_raw(1);
+
+        assert_eq!(
+            xpath!("/proc/1/exe").check(
+                pid,
+                Some(&FileType::Lnk),
+                None,
+                XPathCheckFlags::RESTRICT_MAGICLINKS
+            ),
+            Err(Errno::ENOENT),
+        );
+        assert_eq!(
+            xpath!("/proc/1/exe").check(
+                pid,
+                Some(&FileType::MagicLnk(ProcMagic::Exe { pid })),
+                None,
+                XPathCheckFlags::RESTRICT_MAGICLINKS
+            ),
+            Err(Errno::ENOENT),
+        );
+        assert_eq!(
+            xpath!("/proc/self").check(
+                pid,
+                Some(&FileType::Lnk),
+                None,
+                XPathCheckFlags::RESTRICT_MAGICLINKS
+            ),
+            Ok(()),
+        );
+        assert_eq!(
+            xpath!("/proc/thread-self").check(
+                pid,
+                Some(&FileType::Lnk),
+                None,
+                XPathCheckFlags::RESTRICT_MAGICLINKS
+            ),
+            Ok(()),
+        );
+        assert_eq!(
+            xpath!("/proc/1").check(
+                pid,
+                Some(&FileType::Dir),
+                None,
+                XPathCheckFlags::RESTRICT_MAGICLINKS
+            ),
+            Ok(()),
+        );
+        assert_eq!(
+            xpath!("/tmp/link").check(
+                pid,
+                Some(&FileType::Lnk),
+                None,
+                XPathCheckFlags::RESTRICT_MAGICLINKS
+            ),
+            Ok(()),
+        );
+        assert_eq!(
+            xpath!("/proc/1/exe").check(pid, Some(&FileType::Lnk), None, XPathCheckFlags::empty()),
+            Ok(()),
+        );
+    }
+
     #[test]
     fn test_path_split_prefix_absolute() {
         let path = XPathBuf::from("/tmp/foo/bar/baz");
@@ -2524,7 +2909,7 @@ mod tests {
 
     // Helper to build a buffer.
     fn xp(bytes: &[u8]) -> XPathBuf {
-        XPathBuf(bytes.to_vec())
+        XPathBuf(TinyVec::Heap(bytes.to_vec()))
     }
 
     #[test]
@@ -2535,6 +2920,25 @@ mod tests {
         Ok(())
     }
 
+    #[test]
+    fn test_push_bytes_and_convert_to_vec() {
+        let mut base = XPathBuf::from("./");
+        assert_eq!(base.as_bytes(), b"./");
+        base.append_bytes(b"sigpipe.sock");
+        base.append_byte(0);
+        let v = base.into_vec();
+        assert_eq!(v, b"./sigpipe.sock\0");
+    }
+
+    #[test]
+    fn test_convert_from_string() {
+        let s = String::from("test/path");
+        let xb = XPathBuf::from(s);
+        assert_eq!(xb.as_bytes(), b"test/path");
+        let v = xb.into_vec();
+        assert_eq!(v, b"test/path");
+    }
+
     #[test]
     fn test_replace_prefix_equal_length_overwrite() -> Result<(), Errno> {
         let mut x = xp(b"abc:/tail");
