@@ -2,7 +2,7 @@
 // Syd: rock-solid application kernel
 // src/seal.rs: Execute program as sealed anonymous file
 //
-// Copyright (c) 2024, 2025 Ali Polatel <alip@chesswob.org>
+// Copyright (c) 2024, 2025, 2026 Ali Polatel <alip@chesswob.org>
 // Based in part upon pentacle which is:
 //   Copyright (c) iliana destroyer of worlds <iliana@buttslol.net>
 //   SPDX-License-Identifier: MIT
@@ -17,29 +17,37 @@
 // Commit:e606ab250e6655865bb93a6d98157093f2eb455f
 
 use std::{
-    ffi::CStr,
+    convert::Infallible,
+    env,
+    ffi::{CStr, CString},
     fs::{File, Permissions},
-    ops::{Deref, DerefMut},
-    os::unix::{fs::PermissionsExt as _, io::AsRawFd, process::CommandExt},
-    process::Command,
+    os::unix::{ffi::OsStringExt, fs::PermissionsExt},
 };
 
+use libc::mode_t;
 use nix::{
     errno::Errno,
-    fcntl::{fcntl, FcntlArg, SealFlag},
+    fcntl::{fcntl, openat, AtFlags, FcntlArg, OFlag, SealFlag},
     libc::{
         c_int, c_uint, F_SEAL_FUTURE_WRITE, F_SEAL_GROW, F_SEAL_SEAL, F_SEAL_SHRINK, F_SEAL_WRITE,
         MFD_ALLOW_SEALING, MFD_CLOEXEC, MFD_EXEC, MFD_NOEXEC_SEAL,
     },
+    sys::stat::Mode,
+    unistd::{execveat, Pid},
 };
 
-use crate::{compat::MFdFlags, config::ENV_SKIP_SCMP, err::err2no, io::ReadFd, path::XPathBuf};
+use crate::{
+    compat::{fstatx, MFdFlags, STATX_TYPE},
+    config::ENV_SKIP_SCMP,
+    err::err2no,
+    io::ReadFd,
+    lookup::FileType,
+    path::XPathBuf,
+    proc::proc_open,
+    retry::retry_on_eintr,
+};
 
-// SAFETY: The provided slice is nul-terminated and does not contain any interior nul bytes. On Rust
-// 1.64 and later (rust-lang/rust#99977), these required invariants are checked at compile time.
-//
-// The ideal alternative here is to use C-string literals, introduced in Rust 1.77, but that is a
-// steep MSRV bump to introduce at time of writing this comment.
+// Default memory fd name.
 const DEFAULT_MEMFD_NAME: &CStr = c"syd";
 
 // not yet present in the libc crate
@@ -72,89 +80,99 @@ const OPTIONS: SealOptions = SealOptions::new().close_on_exec(true).executable(t
 /// fails, the `fcntl(2)` `F_GET_SEALS` or `F_ADD_SEALS` commands fail,
 /// or copying from executable file to the anonymous file fails.
 pub fn ensure_sealed() -> Result<(), Errno> {
-    // SAFETY: This runs before sandboxing starts,
-    // so safe open API is not available yet.
+    // Open procfs safely and validate.
+    let fd_proc = proc_open()?;
+
+    // Open $pid/exe relative to the validated procfs.
+    // SAFETY: Use $pid rather than self to reduce determinism.
+    let mut pexe = XPathBuf::from_pid(Pid::this())?;
+    pexe.push(b"exe");
+
     #[expect(clippy::disallowed_methods)]
-    let mut file = File::open("/proc/self/exe").map_err(|err| err2no(&err))?;
+    let mut file = openat(
+        fd_proc,
+        &pexe,
+        OFlag::O_RDONLY | OFlag::O_CLOEXEC,
+        Mode::empty(),
+    )
+    .map(File::from)?;
+
     if OPTIONS.is_sealed(&file) {
+        // Already sealed, move on...
         Ok(())
     } else {
-        let mut command = SealedCommand::new(&mut file)?;
-        let mut args = std::env::args_os().fuse();
-        if let Some(arg0) = args.next() {
-            command.arg0(arg0);
-        }
-        command.args(args);
-
-        // Force RUST_BACKTRACE environment variable to 0 for Syd.
-        // Passthrough the original value to the sandbox process.
-        // See syd.rs for the other branch.
-        // Rest is handled in unshare/child.rs.
-        match std::env::var_os("RUST_BACKTRACE") {
-            Some(val) => command.env("SYD_RUST_BACKTRACE", val),
-            None => command.env_remove("SYD_RUST_BACKTRACE"),
-        };
-        if std::env::var_os(ENV_SKIP_SCMP).is_none() {
-            command.env("RUST_BACKTRACE", "0");
-        }
-
-        Err(err2no(&command.exec()))
+        // Copy into memfd, seal and reexec.
+        Err(SealedCommand::new(&mut file)?.exec().unwrap_err())
     }
 }
 
 /// A [`Command`] wrapper that spawns sealed memory-backed programs.
-///
-/// You can use the standard [`Command`] builder methods (such as [`spawn`][`Command::spawn`] and
-/// [`CommandExt::exec`]) via [`Deref` coercion][`DerefMut`].
 pub struct SealedCommand {
-    inner: Command,
-    // we need to keep this memfd open for the lifetime of this struct
-    _memfd: File,
+    memfd: File,
 }
 
 impl SealedCommand {
-    /// Constructs a new [`Command`] for launching the program data in `program` as a sealed
-    /// memory-backed file, with the same default configuration as [`Command::new`].
+    /// Constructs a new [`SealedCommand`] for launching the program
+    /// data in `program` as a sealed memory-backed file.
     ///
-    /// The memory-backed file will close on `execve(2)`.
-    ///
-    /// `argv[0]` of the program will default to "syd". [`CommandExt::arg0`] can override this.
+    /// The memory-backed file will close on `execveat(2)`.
     ///
     /// # Errors
     ///
-    /// An error is returned if `memfd_create(2)` fails, the `fcntl(2)` `F_GET_SEALS` or
-    /// `F_ADD_SEALS` commands fail, or copying from `program` to the anonymous file fails.
+    /// An error is returned if `program` is not a regular file,
+    /// `memfd_create(2)` fails, the `fcntl(2)` `F_GET_SEALS` or
+    /// `F_ADD_SEALS` commands fail, or copying from `program` to the
+    /// anonymous file fails.
     pub fn new<Fd>(program: &mut Fd) -> Result<Self, Errno>
     where
         Fd: ReadFd,
     {
+        // Check the file type and bail if it's not a regular file.
+        let statx = retry_on_eintr(|| fstatx(&program, STATX_TYPE))?;
+        let ftype = FileType::from(mode_t::from(statx.stx_mode));
+        if !ftype.is_file() {
+            return Err(Errno::ENOEXEC);
+        }
+
         let mut memfd = OPTIONS.create()?;
         crate::io::copy(program, &mut memfd)?;
         OPTIONS.seal(&mut memfd)?;
 
-        let mut comm = XPathBuf::from("/proc/thread-self/fd");
-        comm.push_fd(memfd.as_raw_fd());
-        let mut comm = Command::new(comm);
-        comm.arg0("syd");
-
-        Ok(Self {
-            inner: comm,
-            _memfd: memfd,
-        })
+        Ok(Self { memfd })
     }
-}
 
-impl Deref for SealedCommand {
-    type Target = Command;
+    /// Execute the memory-backed file with execveat(2) and AT_EMPTY_PATH.
+    ///
+    /// The file will be closed on execveat(2).
+    pub fn exec(self) -> Result<Infallible, Errno> {
+        // Force RUST_BACKTRACE environment variable to 0 for Syd.
+        // Passthrough the original value to the sandbox process.
+        // See syd.rs for the other branch.
+        // Rest is handled in unshare/child.rs.
+        match env::var_os("RUST_BACKTRACE") {
+            Some(val) => env::set_var("SYD_RUST_BACKTRACE", val),
+            None => env::remove_var("SYD_RUST_BACKTRACE"),
+        };
+        if env::var_os(ENV_SKIP_SCMP).is_none() {
+            env::set_var("RUST_BACKTRACE", "0");
+        }
 
-    fn deref(&self) -> &Command {
-        &self.inner
-    }
-}
+        // Collect arguments.
+        let args = env::args_os()
+            .map(|arg| CString::new(arg.into_vec()).or(Err(Errno::EINVAL)))
+            .collect::<Result<Vec<CString>, Errno>>()?;
 
-impl DerefMut for SealedCommand {
-    fn deref_mut(&mut self) -> &mut Command {
-        &mut self.inner
+        // Collect environment variables.
+        let envs = env::vars_os()
+            .map(|(k, v)| {
+                let mut bytes = k.into_vec();
+                bytes.push(b'=');
+                bytes.extend(v.into_vec());
+                CString::new(bytes).or(Err(Errno::EINVAL))
+            })
+            .collect::<Result<Vec<CString>, Errno>>()?;
+
+        execveat(self.memfd, c"", &args, &envs, AtFlags::AT_EMPTY_PATH)
     }
 }
 
@@ -415,7 +433,7 @@ impl SealOptions {
         };
 
         if self.is_executable_set() {
-            let permissions = file.metadata().or(Err(Errno::EACCES))?.permissions();
+            let permissions = file.metadata().map_err(|err| err2no(&err))?.permissions();
             let new_permissions =
                 Permissions::from_mode(if self.memfd_flags & MFD_NOEXEC_SEAL != 0 {
                     permissions.mode() & !0o111
@@ -426,7 +444,7 @@ impl SealOptions {
                 });
             if permissions != new_permissions {
                 file.set_permissions(new_permissions)
-                    .or(Err(Errno::EACCES))?;
+                    .map_err(|err| err2no(&err))?;
             }
         }
 
