@@ -2,7 +2,7 @@
 // Syd: rock-solid application kernel
 // src/wordexp.rs: Interface for libc's wordexp(3).
 //
-// Copyright (c) 2024, 2025 Ali Polatel <alip@chesswob.org>
+// Copyright (c) 2024, 2025, 2026 Ali Polatel <alip@chesswob.org>
 //
 // SPDX-License-Identifier: GPL-3.0
 
@@ -18,44 +18,51 @@ use std::{
         fd::{AsFd, AsRawFd, BorrowedFd, FromRawFd},
         unix::ffi::OsStrExt,
     },
-    time::{Duration, Instant},
+    time::Instant,
 };
 
 use bitflags::bitflags;
 use data_encoding::HEXLOWER;
+use dur::Duration;
 use libseccomp::{ScmpAction, ScmpFilterContext, ScmpSyscall};
 use memchr::memchr3;
 use nix::{
     errno::Errno,
-    fcntl::OFlag,
+    fcntl::{open, OFlag},
     libc::{_exit, c_char, size_t, CLONE_FILES, ENOSYS, SIGCHLD, SIGKILL, SIGSYS},
-    mount::{mount, MsFlags},
+    mount::MsFlags,
     sched::{unshare, CloneFlags},
     sys::{
-        prctl::set_pdeathsig,
+        resource::Resource,
         signal::{sigprocmask, SigSet, SigmaskHow, Signal},
+        stat::Mode,
         wait::{Id, WaitPidFlag},
     },
-    unistd::{chdir, close, Gid, Pid, Uid},
+    unistd::{chdir, Gid, Pid, Uid},
 };
 
 use crate::{
-    compat::{pipe2_raw, waitid, MFdFlags, WaitStatus},
+    compat::{pipe2_raw, set_pdeathsig, waitid, MFdFlags, WaitStatus},
     config::{MINI_STACK_SIZE, *},
-    confine::{confine_mdwe, CLONE_NEWTIME, MS_NOSYMFOLLOW},
+    confine::{confine_mdwe, confine_rlimit, CLONE_NEWTIME},
     cookie::safe_memfd_create,
     debug,
     err::err2no,
-    fs::{pidfd_send_signal, safe_clone, seal_memfd_all, set_cloexec, set_nonblock},
+    fd::{close, pidfd_send_signal, seal_memfd_all, set_cloexec, set_nonblock},
+    fs::safe_clone,
     get_user_home, get_user_name,
     hash::SydHashSet,
     landlock::RulesetStatus,
     landlock_policy::LandlockPolicy,
     log::contains_ascii_unprintable,
     lookup::safe_copy_if_exists,
+    mount::{
+        api::MountAttrFlags,
+        util::{mount_bind, mount_fs, set_root_mount_propagation},
+    },
     path::PATH_MAX,
-    unshare::{GidMap, UidMap},
-    xpath, XPath, XPathBuf,
+    proc::{proc_map_user, proc_open},
+    xpath, XPathBuf,
 };
 
 bitflags! {
@@ -102,7 +109,7 @@ pub enum WordExpError {
     /// Process was aborted unexpectedly with signal.
     ProcessError(i32),
     /// Timeout error
-    TimeoutError(u64),
+    TimeoutError(u128),
 }
 
 /// Out of memory.
@@ -441,7 +448,7 @@ impl WordExp<'_> {
         let mut ret = Vec::new();
 
         loop {
-            if !sig && (err as i32 != 0 || epoch.elapsed() >= timeout) {
+            if !sig && (err as i32 != 0 || epoch.elapsed() >= timeout.into()) {
                 // a. Out of memory condition
                 // b. Timeout exceeded
                 // Send SIGKILL once, and fall-through to wait.
@@ -467,7 +474,7 @@ impl WordExp<'_> {
                         continue;
                     }
                     Err(ref e) if matches!(err2no(e), Errno::EAGAIN | Errno::EINTR) => {
-                        std::thread::sleep(Duration::from_millis(100));
+                        std::thread::sleep(Duration::from_millis(100).into());
                         continue;
                     }
                     Err(ref e) => {
@@ -554,6 +561,10 @@ impl WordExp<'_> {
     #[expect(clippy::cognitive_complexity)]
     #[expect(clippy::disallowed_methods)]
     pub fn confine() {
+        if env::var_os(ENV_SKIP_SCMP).is_some() {
+            return;
+        }
+
         // SAFETY: Determine user HOME directory.
         // This will be confined by Landlock.
         let uid = Uid::current();
@@ -571,104 +582,7 @@ impl WordExp<'_> {
 
         // SAFETY: set up namespace isolation.
         // continue on errors as unprivileged userns may not be supported.
-        if unshare(
-            CloneFlags::CLONE_NEWUSER
-                | CloneFlags::CLONE_NEWCGROUP
-                | CloneFlags::CLONE_NEWIPC
-                | CloneFlags::CLONE_NEWNET
-                | CloneFlags::CLONE_NEWNS
-                | CloneFlags::CLONE_NEWPID
-                | CloneFlags::CLONE_NEWUTS
-                | CLONE_NEWTIME,
-        )
-        .is_ok()
-        {
-            debug!("ctx": "expand",
-                "msg": "created and entered into new user, mount, pid, network, cgroup, ipc, uts, and time namespaces");
-            // SAFETY: map current user in new user namespace.
-            let uid_buf = {
-                let uid_maps = vec![
-                    UidMap {
-                        inside_uid: uid.into(),
-                        outside_uid: uid.into(),
-                        count: 1,
-                    }, // Map the current user.
-                ];
-                let mut buf = Vec::new();
-                for map in uid_maps {
-                    writeln!(
-                        &mut buf,
-                        "{} {} {}",
-                        map.inside_uid, map.outside_uid, map.count
-                    )
-                    .expect("write uid_map");
-                }
-                buf
-            };
-            let gid_buf = {
-                let gid_maps = vec![
-                    GidMap {
-                        inside_gid: gid.into(),
-                        outside_gid: gid.into(),
-                        count: 1,
-                    }, // Map the current group.
-                ];
-                let mut buf = Vec::new();
-                for map in gid_maps {
-                    writeln!(
-                        &mut buf,
-                        "{} {} {}",
-                        map.inside_gid, map.outside_gid, map.count
-                    )
-                    .expect("write gid_map");
-                }
-                buf
-            };
-            // Write "deny" to /proc/self/setgroups before writing to gid_map.
-            File::create("/proc/self/setgroups")
-                .and_then(|mut f| f.write_all(b"deny"))
-                .expect("deny setgroups");
-            File::create("/proc/self/gid_map")
-                .and_then(|mut f| f.write_all(&gid_buf[..]))
-                .expect("map current group");
-            debug!("ctx": "expand",
-                "msg": format!("mapped current group {gid} into new user namespace"));
-            File::create("/proc/self/uid_map")
-                .and_then(|mut f| f.write_all(&uid_buf[..]))
-                .expect("map current user");
-            debug!("ctx": "expand",
-                "msg": format!("mapped current user {uid} into new user namespace"));
-
-            // SAFETY: remount rootfs as readonly,nosuid,nodev,nosymfollow
-            // Careful here, unshare(2) may be available but mount(2) may not be,
-            // so we must handle mount errors gracefully.
-            let mut flags: MsFlags = MsFlags::MS_BIND
-                | MsFlags::MS_REC
-                | MsFlags::MS_RDONLY
-                | MsFlags::MS_NOSUID
-                | MsFlags::MS_NODEV
-                | MS_NOSYMFOLLOW;
-            if mount(
-                Some("none"),
-                "/",
-                None::<&XPath>,
-                MsFlags::MS_PRIVATE | MsFlags::MS_REC,
-                None::<&XPath>,
-            )
-            .is_ok()
-            {
-                debug!("ctx": "expand",
-                    "msg": "set mount propagation to private in the new mount namespace");
-                mount(Some("/"), "/", Some("/"), flags, None::<&XPath>).expect("remount rootfs");
-                debug!("ctx": "expand",
-                    "msg": "remounted root with readonly, nosuid, nodev, and nosymfollow options in the new mount namespace");
-                // SAFETY: mount private procfs
-                // pid=1 is required to exist before this.
-                flags.remove(MsFlags::MS_BIND | MsFlags::MS_REC | MS_NOSYMFOLLOW);
-                flags.insert(MsFlags::MS_NOEXEC);
-                Self::mount_proc(flags);
-            }
-        }
+        let _ = Self::setup_namespaces(uid, gid);
 
         // SAFETY: Landlock: confine filesystem as read-only.
         // continue on errors as Landlock may not be supported.
@@ -829,8 +743,11 @@ impl WordExp<'_> {
             }
         }
 
-        // SAFETY: confine with seccomp.
-        // panics on errors.
+        // Set file size rlimits to zero, panic on errors.
+        #[expect(clippy::disallowed_methods)]
+        confine_rlimit(Resource::RLIMIT_FSIZE, None).expect("setrlimit(RLIMIT_FSIZE,0)");
+
+        // SAFETY: confine with seccomp, panics on errors.
         Self::confine_seccomp();
     }
 
@@ -852,7 +769,8 @@ impl WordExp<'_> {
 
         for sysname in WORDEXP_SYSCALLS
             .iter()
-            .chain(GET_ID_SYSCALLS)
+            .chain(FUTEX_SYSCALLS)
+            .chain(GETID_SYSCALLS)
             .chain(VDSO_SYSCALLS)
         {
             if let Ok(syscall) = ScmpSyscall::from_name(sysname) {
@@ -869,13 +787,62 @@ impl WordExp<'_> {
 
     #[expect(clippy::cognitive_complexity)]
     #[expect(clippy::disallowed_methods)]
-    fn mount_proc(flags: MsFlags) {
+    fn setup_namespaces(uid: Uid, gid: Gid) -> Result<(), Errno> {
+        unshare(
+            CloneFlags::CLONE_NEWUSER
+                | CloneFlags::CLONE_NEWCGROUP
+                | CloneFlags::CLONE_NEWIPC
+                | CloneFlags::CLONE_NEWNET
+                | CloneFlags::CLONE_NEWNS
+                | CloneFlags::CLONE_NEWPID
+                | CloneFlags::CLONE_NEWUTS
+                | CLONE_NEWTIME,
+        )?;
+        debug!("ctx": "expand",
+            "msg": "created and entered into new user, mount, pid, network, cgroup, ipc, uts, and time namespaces");
+
+        // Set up UID/GID mapping in new user namespace.
+        proc_map_user(proc_open()?, uid, gid, false /*map_root*/)?;
+
+        // SAFETY: Remount rootfs as readonly,nosuid,nodev,nosymfollow.
+        let mut flags = MountAttrFlags::MOUNT_ATTR_RDONLY
+            | MountAttrFlags::MOUNT_ATTR_NOSUID
+            | MountAttrFlags::MOUNT_ATTR_NODEV
+            | MountAttrFlags::MOUNT_ATTR_NOSYMFOLLOW;
+
+        // Set mount propagation to private.
+        set_root_mount_propagation(MsFlags::MS_PRIVATE)?;
+        debug!("ctx": "expand",
+            "msg": "set mount propagation to private in new mount namespace");
+
+        // Remount root.
+        open(
+            "/",
+            OFlag::O_CLOEXEC | OFlag::O_PATH | OFlag::O_DIRECTORY | OFlag::O_NOFOLLOW,
+            Mode::empty(),
+        )
+        .and_then(|root| mount_bind(&root, &root, flags))?;
+        debug!("ctx": "expand",
+            "msg": "remounted root with readonly, nosuid, nodev, and nosymfollow options in new mount namespace");
+
+        // SAFETY: Mount private procfs.
+        // pid=1 is required to exist before this.
+        flags.remove(MountAttrFlags::MOUNT_ATTR_NOSYMFOLLOW);
+        flags.insert(MountAttrFlags::MOUNT_ATTR_NOEXEC);
+        Self::mount_proc(flags);
+
+        Ok(())
+    }
+
+    #[expect(clippy::cognitive_complexity)]
+    #[expect(clippy::disallowed_methods)]
+    fn mount_proc(flags: MountAttrFlags) {
         let mut stack = [0u8; MINI_STACK_SIZE];
         safe_clone(
             Box::new(move || -> isize {
                 // pid=1 here.
                 debug!("ctx": "expand",
-                    "msg": "started init process in the new pid namespace");
+                    "msg": "started init process in new pid namespace");
 
                 // SAFETY: set parent-death signal to SIGKILL
                 if set_pdeathsig(Some(Signal::SIGKILL)).is_err() {
@@ -889,16 +856,22 @@ impl WordExp<'_> {
                     .expect("block signals");
 
                 // SAFETY: mount private procfs, continue on errors.
-                match mount(
-                    Some("proc"),
+                match open(
                     "/proc",
-                    Some("proc"),
-                    flags,
-                    Some("hidepid=4"),
-                ) {
+                    OFlag::O_CLOEXEC | OFlag::O_PATH | OFlag::O_DIRECTORY,
+                    Mode::empty(),
+                )
+                .and_then(|proc| {
+                    mount_fs(
+                        OsStr::new("proc"),
+                        proc,
+                        flags,
+                        Some("hidepid=4,subset=pid"),
+                    )
+                }) {
                     Ok(_) => {
                         debug!("ctx": "expand",
-                            "msg": "mounted proc with hidepid=4 in the new mount namespace");
+                            "msg": "mounted proc with hidepid=4,subset=pid in new mount namespace");
                     }
                     Err(errno) => {
                         debug!("ctx": "expand",
@@ -907,7 +880,7 @@ impl WordExp<'_> {
                 };
 
                 // SAFETY: block until the parent-death signal kills us.
-                std::thread::sleep(Duration::MAX);
+                std::thread::sleep(std::time::Duration::MAX);
 
                 unreachable!();
             }),
