@@ -28,7 +28,10 @@ use crate::{
     alert,
     cache::SysInterrupt,
     config::*,
-    confine::{confine_scmp_madvise, confine_scmp_write, scmp_add_setid_rules, ExportMode},
+    confine::{
+        confine_scmp_fcntl, confine_scmp_madvise, confine_scmp_prctl, confine_scmp_write,
+        confine_scmp_wx_syd, scmp_add_setid_rules, ExportMode,
+    },
     cookie::{CookieIdx, SYSCOOKIE_POOL},
     err::{err2no, scmp2no, SydJoinHandle, SydResult},
     error,
@@ -521,49 +524,13 @@ impl Interrupter {
         }
 
         // Allow safe fcntl(2) utility calls.
-        for sysname in ["fcntl", "fcntl64"] {
-            let syscall = match ScmpSyscall::from_name(sysname) {
-                Ok(syscall) => syscall,
-                Err(_) => {
-                    info!("ctx": "confine", "op": "allow_int_syscall",
-                        "msg": format!("invalid or unsupported syscall {sysname}"));
-                    continue;
-                }
-            };
-
-            for op in INT_FCNTL_OPS {
-                ctx.add_rule_conditional(ScmpAction::Allow, syscall, &[scmp_cmp!($arg1 == *op)])?;
-            }
-        }
+        confine_scmp_fcntl(&mut ctx, INT_FCNTL_OPS)?;
 
         // Allow safe prctl(2) operations.
-        let sysname = "prctl";
-        if let Ok(syscall) = ScmpSyscall::from_name(sysname) {
-            for (_, op) in INT_PRCTL_OPS {
-                ctx.add_rule_conditional(ScmpAction::Allow, syscall, &[scmp_cmp!($arg0 == *op)])?;
-            }
-        } else {
-            info!("ctx": "confine", "op": "allow_int_syscall",
-                "msg": format!("invalid or unsupported syscall {sysname}"));
-        }
+        confine_scmp_prctl(&mut ctx, INT_PRCTL_OPS)?;
 
         // Prevent executable memory.
-        const PROT_EXEC: u64 = libc::PROT_EXEC as u64;
-        for sysname in ["mmap", "mmap2", "mprotect"] {
-            match ScmpSyscall::from_name(sysname) {
-                Ok(syscall) => {
-                    ctx.add_rule_conditional(
-                        ScmpAction::Allow,
-                        syscall,
-                        &[scmp_cmp!($arg2 & PROT_EXEC == 0)],
-                    )?;
-                }
-                Err(_) => {
-                    info!("ctx": "confine", "op": "allow_int_syscall",
-                        "msg": format!("invalid or unsupported syscall {sysname}"));
-                }
-            }
-        }
+        confine_scmp_wx_syd(&mut ctx)?;
 
         // Allow writes to the log-fd.
         // No proc_pid_mem(5) access required here.
@@ -576,7 +543,7 @@ impl Interrupter {
         for sysname in INT_SYSCALLS
             .iter()
             .chain(FUTEX_SYSCALLS)
-            .chain(GET_ID_SYSCALLS)
+            .chain(GETID_SYSCALLS)
             .chain(VDSO_SYSCALLS)
         {
             match ScmpSyscall::from_name(sysname) {
